"""This module implements core functions of the
IRT Graded Response Model (Samejima, 1997).

These functions are collected in a separate module to simplify
selecting either this model, or any other model, in future versions.

*** Main Class: GradedResponse

*** References:
F. Samejima (1997). Graded response model.
In W. J. v. D. Linden and R. K. Hambleton, eds.,
*Handbook of Modern Item Response Theory*, p. 85–100. Springer, New York.

J.-P. Fox (2010). *Bayesian Item Response Modeling: Theory and Applications*.
Statistics for Social and Behavioral Sciences. Springer.

*** Version:
* Version 1.0.0:
2023-07-22, tested gradient function OK with scipy.check_grad
2023-07-21, modified methods for jointly sampled (eta, theta) parameters

2020-06-14, first version with methods copied from item_respondents
    to replace functions previously defined there.
2020-06-15, cleanup, tested
"""
import numpy as np

from ItemResponseCalc.ir_latent import LatentLogistic

import logging
logger = logging.getLogger(__name__)


# --------------------------------------------------------------------------
class GradedResponse:
    """Implementing central methods to calculate
    log probability of observed responses, and corresponding gradients,
    given item and respondent parameters.
    """
    def __init__(self, latent_class=LatentLogistic, scale=None):
        """
        :param latent_class: class of latent variable that determines responses
        :param scale: (optional) 1D array of scale factors
            to get same response probabilities after external change of (eta, theta) scale
        """
        self.latent_class = latent_class
        self.scale = scale  # *** not needed not used in this version?

    def latent_scale(self):
        return self.latent_class.scale

    def log_response_prob(self, tau, theta):
        """Probability-mass distribution over all possible responses
        for given parameters.
        :param tau: 1D or 2D array of all response thresholds, INCL. extremes at -inf, +inf
            tau[..., m, l] = (..., m)-th sample of lower threshold for Response = l
        :param theta: 1D array with hypothetical latent-variable locations
        :return: p = 2D or 3D array with
            p[..., t, l] = prob{l-th response | ...-th tau vector, t-th theta value}
        """
        lp = np.array([self.latent_class.log_cdf_diff(tau[..., l, None] - theta,
                                                      tau[..., l + 1, None] - theta)
                       for l in range(tau.shape[-1] - 1)])
        return np.moveaxis(lp, 0, -1)

    def d_log_response_prob_d_theta(self, tau, theta):  # ******** not tested !
        """Partial derivative of log_response_prob(self, tau, theta) w.r.t. theta.
         :param tau: 1D or 2D array of all response thresholds, INCL. extremes at -inf, +inf
             tau[..., m, l] = (..., m)-th sample of lower threshold for Response = l
         :param theta: 1D array with hypothetical latent-variable locations
         :return: dlp = 2D or 3D array with
             dlp[..., t, l] = d ln P{response = l | tau[..., :], theta[t]} / d th[t]
         """
        dlp = np.array([-sum(self.latent_class.d_log_cdf_diff(tau[..., l, None] - theta,
                                                             tau[..., l + 1, None] - theta))
                        for l in range(tau.shape[-1] - 1)])
        return np.moveaxis(dlp, 0, -1)

    def response_prob(self, tau, theta):
        """Probability-mass distribution over all possible responses
        for given parameters.
        :param tau: 1D or 2D array of all response thresholds, INCL. extremes at -inf, +inf
        :param theta: 1D array with hypothetical latent-variable locations
        :return: p = 2D or 3D array with
            p[..., t, l] = prob{l-th response | ...-th tau vector, t-th theta value}
        """
        return np.exp(self.log_response_prob(tau, theta))

    # def item_response_logprob(self, r, tau, theta):  # *** not needed ?
    #     """Log Probability of responses to ONE item, given parameters, summed across respondents
    #     :param r: 1D array with r[s] = ordinal response by s-th respondent to this item
    #         valid responses in range(L_i), missing response = -1
    #         L_i = number of response categories
    #         len(r) == theta.shape[1]
    #     :param tau: tau[n, :] = n-th sample of response interval limits
    #         INCL. all elements in [-inf, +inf]
    #         tau.shape[-1] == L_i + 1
    #     :param theta: theta[n, s, d] = n-th sample of d-th trait for s-th subject
    #         theta.shape[0] == tau.shape[0] == number of joint samples
    #         theta.shape[1] == len(r)
    #     :return: lp = 3D array with elements
    #         lp[n, s, d] = ln P(r[s] | tau[n, :], theta[n, s, d])
    #     """
    #     tau_low, tau_high = _response_interval(tau, r)
    #     a = tau_low[..., None] - theta
    #     b = tau_high[..., None] - theta
    #     # a[n, s, d] = n-th sample, s-th respondent, d-th trait; b[n, s, d] same
    #     return self.latent_class.log_cdf_diff(a, b)

    def item_sum_response_logprob(self, r, tau, theta):
        """Log Probability of responses to ONE item, given parameters,
        summed across all respondents
        :param r: 1D array with r[s] = ordinal response by s-th subject to this item
            valid responses in range(L_i), missing response = -1
            L_i = number of response categories
            len(r) == theta.shape[1]
        :param tau: 2D array with tau[n, :] = n-th sample of response interval limits
            INCL. all elements in [-inf, +inf]
            tau.shape[-1] == L_i + 1
        :param theta: 3D array with theta[n, s, d] = n-th sample of d-th trait for s-th subject
            theta.shape[0] == tau.shape[0] == number of joint samples
            theta.shape[1] == len(r)
        :return: lp = 2D array with elements
            lp[n, d] = sum_s ln P(r[s] | tau[n, :], theta[n, s, d])
        """
        tau_low, tau_high = _response_interval(tau, r)
        a = tau_low[..., None] - theta
        b = tau_high[..., None] - theta
        # a[n, s, d] = n-th sample, s-th respondent, d-th trait
        # b[n, s, d] similar
        return np.sum(self.latent_class.log_cdf_diff(a, b),
                      axis=1)

    def d_item_sum_response_logprob(self, r, tau, theta):
        """Gradients of item_sum_response_logprob w.r.t tau and theta, for ONE item
        :param r: 1D array with r[s] = ordinal response by s-th subject to this item
            valid responses in range(L_i), missing response = -1
            L_i = number of response categories
            len(r) == theta.shape[1]
        :param tau: tau[n, :] = n-th sample of response interval limits
            INCL. all elements in [-inf, +inf]
            tau.shape[-1] == L_i + 1
        :param theta: theta[n, s, d] = n-th sample of d-th trait for s-th subject
            theta.shape[0] == tau.shape[0] == number of joint samples
            theta.shape[1] == len(r)
        :return: tuple (dlp_tau, dlp_theta), with elements
            dlp_tau[n, k, d] = sum_s d ln P(r[s] | tau[n, :], theta[n, s, d]) / d tau[n, k]
            dlp_theta[n, s, d] = d ln P(r[s] | tau[n, :], theta[n, s, d]) / d theta[n, s, d]
            NOTE: dlp_tau[:, 0, :] == dlp_tau[:, -1, :] == 0.
        """
        tau_low, tau_high = _response_interval(tau, r)
        a = tau_low[..., None] - theta
        b = tau_high[..., None] - theta
        # a[n, s, d] = n-th sample, s-th respondent, d-th trait; b[n, s, d] same
        dll_da, dll_db = self.latent_class.d_log_cdf_diff(a, b)
        # dll_da[n, s, d] = d log P(r[s] | tau[n, :], theta[n, s, d]) / d a[n, s, d]
        # dll_db[n, s, d] = d log P(r[s] | tau[n, :], theta[n, s, d]) / d b[n, s, d]
        dll_tau = np.zeros(tau.shape + theta.shape[-1:])
        for l in range(1, tau.shape[-1] - 1):  # each internal threshold
            dll_tau[:, l, :] += np.sum(dll_da[:, r == l, :], axis=1)
            dll_tau[:, l, :] += np.sum(dll_db[:, r + 1 == l, :], axis=1)
        dll_theta = - dll_da - dll_db
        return dll_tau, dll_theta


# -------------------------------- help functions
def _response_interval(tau, r):
    """Get response interval for given thresholds and responses for ONE item
    :param tau: 2D array with threshold samples, as
        tau[m, l] = m-th sample of LOWER limit of l-th response interval,
            INCL extreme limits -inf, +inf
        tau.shape[-1] == L + 1, with L = number of response levels
    :param r: 1D array of response indices, with missing response = -1
        r[s] = response index for s-th subject in range(L)
        r[s] + 1 <= tau.shape[-1]
        except r[s] = -1 in case of missing response
    :return: tuple (tau_low, tau_high) with
        tau_low[m, s] = m-th sample of lower interval limit for r[s]
        tau_high[m, s] = m-th sample of upper interval limit for r[s]
        tau_low.shape == tau_high.shape
    """
    r = np.asarray(r)
    tau_low = tau[:, np.maximum(0, r)]  # = -inf for r[s] == 0 OR r[s] == -1
    tau_high = tau[:, r + 1]
    tau_high[:, r < 0] = np.inf  # for r[s] == -1
    return tau_low, tau_high


# -------------------------------------------------------- TEST:
if __name__ == '__main__':
    # Testing some module functions
    from ItemResponseCalc.ir_thresholds import ThresholdsFree
    from scipy.optimize import check_grad, approx_fprime

    print('*** Testing _rating_thresholds ***')
    n_categories = 5
    n_samples = 2
    thr = ThresholdsFree()
    n_eta = n_categories - 1
    eta = np.zeros((n_samples, n_eta))
    tau = thr.tau(eta)
    print('tau=\n', np.array_str(tau, precision=3))
    r = list(range(n_categories)) + [-1, 0]
    r = np.array(r)
    tau_low, tau_high = _response_interval(tau, r)
    print('_response_intervals=')
    for (t0, t1) in zip(tau_low.T, tau_high.T):
        print(f'[{t0[0]:.3f}, {t1[0]:.3f}]')

    print('\n*** Check gradients d_item_sum_response_logprob: ')
    grm = GradedResponse()

    n_subjects = len(r)
    n_traits = 3
    test_subject = 0    # subject index for test
    test_theta = 0  # trait index for test
    test_tau = 3    # tau index for test

    theta = np.zeros((n_samples, n_subjects, 1)) + np.arange(n_traits)

    def test_logprob_theta(y):
        # len(y) == n_subjects
        th = theta + y[:, None]
        return grm.item_sum_response_logprob(r, tau, th)[0, test_theta]

    def test_d_logprob_theta(y):
        th = theta + y[:, None]
        dll_dth = grm.d_item_sum_response_logprob(r, tau, th)[1]
        # dll_dth[n, s, d] = d item_sum_response_logprob(...)[n, d] / d th[n, s, d] for all d
        # *** BUT test_logprob_theta returned only d == test_theta
        return dll_dth[0, :, test_theta]

    def test_logprob_tau(y):
        t = np.concatenate(([-np.inf], y, [np.inf]))
        t = np.tile(t, (n_samples, 1))
        return grm.item_sum_response_logprob(r, t, theta)[0, test_theta]

    def test_d_logprob_tau(y):
        t = np.concatenate(([-np.inf], y, [np.inf]))
        t = np.tile(t, (n_samples, 1))
        dll_dt = grm.d_item_sum_response_logprob(r, t, theta)[0]
        return dll_dt[0, 1:-1, test_theta]
    # ------------------------------------

    print('tau=\n', np.array_str(tau, precision=3))
    print('theta=\n', np.array_str(theta, precision=3))
    print('item_sum_response_logprob(r, tau, theta)=',
          grm.item_sum_response_logprob(r, tau, theta))

    theta_y = np.zeros(n_subjects)
    for test_theta in range(n_traits):
        print(f'\n*** Testing d_logprob_theta with theta_y={theta_y} at trait no. {test_theta}')
        print(f'\ntest_logprob_theta(test_y) = {test_logprob_theta(theta_y)}')

        err = check_grad(test_logprob_theta, test_d_logprob_theta, theta_y)
        print('test_d_logprob_theta =', test_d_logprob_theta(theta_y))
        print('approx_grad = ', approx_fprime(theta_y,
                                              test_logprob_theta,
                                              epsilon=1e-6))
        print('check_grad err = ', err)

    for test_tau in range(tau.shape[-1] - 2):
        # tau_y = 0.4  # test value to adjust tau[:, test_tau]
        tau_y = tau[0, 1:-1]  # only internal thresholds
        tau_y[test_tau] += 0.4
        print(f'\n*** Testing d_logprob_tau with tau_y={tau_y} at threshold no. {test_tau}:')
        print(f'\ntest_logprob_tau(tau_y) = {test_logprob_tau(tau_y)}')

        err = check_grad(test_logprob_tau, test_d_logprob_tau, tau_y)
        print('test_d_logprob_tau =', test_d_logprob_tau(tau_y))
        print('approx_grad = ', approx_fprime(tau_y,
                                              test_logprob_tau,
                                              epsilon=1e-6))
        print('check_grad err = ', err)



"""This module defines help classes to access responses to questionnaire items.
Data may be stored in various file formats.

*** Class Overview:

ItemResponseDataSet: reference to all response data for selected group(s),
    to be used as input for statistical analysis.
    Each response source must be an iterable over response records.
    A Response Record is a 1D array-like with ordinal responses from ONE subject,
    with one ordinal index value for each Questionnaire Item.
    The input data for one group may be
    1) a single list of subject response lists,
    2) an ir_source.ItemResponseTable (subclass) instance, defining ONE data source,
    3) an ir_source.Tables instance, chaining input from SEVERAL sources.

Questionnaire: description of a questionnaire, usually with several items

Item: description of a single item (question and answer alternatives)

*** Input Source Formats:
Data may be accessed from simple lists, or from any type of source that Pandas can handle,
e.g., csv or xlsx files, or SQL database(s).
pandas.read_xxx(...), or similar user-defined function, can be used for actual reading.

Each source may have responses encoded in a specific way, different from other sources.
Allowed response alternatives must be specified by user for each input source.

*** Usage Example:

q_file = string or Path identifying questionnaire text (utf8) file
r_file_xxx = string or Path identifying a response file

group0_table = item_response_table(source=r_file0, items=...)
group1_table = item_response_table(r_file1, items=...)
group2A_table = item_response_table(r_file2A, items=...)
group2B_table = item_response_table(r_file2B, items=...)

In each source, items = a dict with elements (column header, list of ordinal response categories)
exactly as encoded in that data source.
Any other value is interpreted as a missing response.
Length of items must agree with number of questionnaire items.

ids = ItemResponseDataSet.load(questionnaire=Questionnaire.load(q_file),
        groups={'Group0': group0_table,
                'Group1': group1_table,
                'Group2', Tables(group2A_table, group2B_table)
                }
        )

This data-set can then be used as input to create an analysis model.
See run_irt.py for a complete example.

An index_col may be defined in each input table for a column with respondent IDs.
Duplicate indices will then be removed on reading, but only within each input data chunk.

*** Version History:
* Version 1.0.0:
2023-09-01, ItemResponseDataSet.response_count() creates ResultTable objects, with save() method

* Version 0.6.0:
2023-07-09, use pandas.read_sql to access database source.
2022-09-15, use pandas.read_xxx functions to access input data files.

* Version 0.5:
2019-07-23, functional version, may need some cleanup
2019-07-29, new version with general table_reader interface
2019-08-25, minor cleanup
"""
# *** item_response_count as TableRef to allow save ***
# *** use subject_ID for individual displays?
# *** use item_ID for item displays ?

import numpy as np
from pathlib import Path
import pandas as pd
from .ir_display_format import ResultTable

import logging

logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------------
class Item:
    """Information describing ONE questionnaire item.
    """
    # *** include optional ID label ? ***
    def __init__(self, question, responses, reverse=False):
        """
        :param question: string with explicit question text
        :param responses: list of ordinal response categories, one for each possible response
            as presented in the questionnaire
        :param reverse: boolean = True if responses should be indexed in reverse order
            NOTE: Response data files may or may not already have recoded
            the responses by ordinal indices in the desired analysis order.
        """
        self.question = question
        self.responses = responses
        self.reverse = reverse

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t'
                + ',\n\t'.join((f'{k}={repr(v)}' for (k, v) in self.__dict__.items()))
                + ')')

    @property
    def n_response_levels(self):
        return len(self.responses)

    @property
    def ordinal_responses(self):
        """Responses in order of increasing underlying trait
        :return: iterator over ordinal response categories (NOT a copy)
        """
        return reversed(self.responses) if self.reverse else self.responses

    def response_codes(self):
        """Response categories as code values in range(n_response_levels)
        :return: list
        """
        return list(range(self.n_response_levels))


class Questionnaire:
    """Container for a questionnaire with several items,
    required to specify number of items and number of response levels for each item,
    even if the actual questions and response alternatives are not included.
    """
    def __init__(self, header=None, items=None, item_response_levels=None):
        """
        :param header: (optional) string describing the questionnaire
        :param items: (optional) list of Item instances
        :param item_response_levels: (optional) used only if items is None:
            list with
            item_response_levels[i] = integer number of response levels for i-th item
        """
        self.header = header
        if items is None:
            self.items = [Item(question=f'Question {i+1}',
                               responses=list(range(1, n_levels +1)))
                          for (i, n_levels) in enumerate(item_response_levels)]
            # self.item_response_levels = item_response_levels
        else:
            self.items = items

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t'
                + ',\n\t'.join((f'{k}={repr(v)}' for (k,v) in self.__dict__.items()))
                + '\n\t)')

    @property
    def n_items(self):
        return len(self.items)

    @classmethod
    def load(cls, file, sep='\t'):
        """Read questions and answers from given file
        :param file: string or Path object identifying input file
        :param sep: (optional) string, separator between response alternatives
        :return: cls instance
        """
        strip = ' \n'  # chars to strip away from end of input lines
        with Path(file).open('rt', encoding='utf8') as f:
            header = f.readline()
            items = []
            while True:
                q = f.readline()
                a = f.readline()
                reverse = f.readline()
                if a == '':
                    break  # even if q is non-empty
                else:
                    items.append(Item(question=q.rstrip(strip),
                                      responses=(a.rstrip(strip)).split(sep),
                                      reverse='rev' in reverse))
        return cls(header.rstrip(strip), items)


# ------------------------------------------------------------
class ItemResponseDataSet:
    """All result data for one complete item response analysis,
    including one or several groups of subjects.
    There should be at least one group with complete individual response data.
    """
    def __init__(self, questionnaire, groups):
        """
        :param questionnaire: single Questionnaire object
        :param groups: dict with elements (g_name, g_subjects), where
            g_name = string identifying the group
            g_subjects = iterable of responses, either,
                (1) a simple list or array,
                (2) OR an ir_source.ItemResponseTable subclass instance,
                (3) OR an ir_source.Tables instance
                    with several join-ed ir_source.ItemResponseTable objects.
            Each input table yields a sequence of response_record, one for each subject.
                response_record = a 1D list with
                response_record[i] = integer ordinal index of response to i-th item
                encoded with origin == 0.
                Missing responses are indicated as -1.
        """
        assert questionnaire is not None, 'Must have a Questionnaire object'
        self.questionnaire = questionnaire
        if groups is None:
            groups = dict()
        self.groups = groups

    def __repr__(self):
        return (self.__class__.__name__ + '('
                + f'\n\t questionnaire= {self.questionnaire.__class__.__name__} '
                + f'object with {self.questionnaire.n_items} items,'
                + '\n\t groups= {\n\t\t'
                + ',\n\t\t'.join(f'{g_key}: {g.__class__.__name__} object with {len(g)} respondent records'
                                 for (g_key, g) in self.groups.items()) + '}'
                + ')')

    @property
    def n_items(self):
        return self.questionnaire.n_items

    @property
    def n_groups(self):
        return len(self.groups)

    def response_counts(self, normalize=False):
        """Tabulate distributions of response counts by items and groups
        :param normalize: (optional) True: present relative frequencies,
            False: absolute response counts
        :return: list of pd.DataFrame objects,  *** TableRef to allow save ? ***
            one count table for each item, each with one column for each group
        """
        res_df = [pd.DataFrame(index=pd.CategoricalIndex(item.ordinal_responses,
                                                         copy=True,
                                                         ordered=True,
                                                         name='Response'),
                               columns=[], dtype=int)  # empty df
                  for item in self.questionnaire.items]
        res_df = [ResultTable(df_i, name=f'Q{i+1}_count')
                  for (i, df_i) in enumerate(res_df)]
        # = space for results
        for (g, g_table) in self.groups.items():
            g_codes = np.array([row for row in g_table])
            for (i, item) in enumerate(self.questionnaire.items):
                gi_series = pd.Series(pd.Categorical.from_codes(g_codes[:, i],
                                                                categories=item.responses,
                                                                ordered=True))
                gi_count = gi_series.value_counts(sort=False,
                                                  normalize=normalize)
                res_df[i][g] = gi_count.reindex(item.responses, fill_value=0)
        return res_df

    def item_response_count(self):
        """Number of response counts for each item, summed across all groups.
        :return: c = list of response count arrays, with
            c[i] = 1D array for i-th item,
            with responses encoded with origin == 0, i.e.,
            c[i][l] = count of l-th ordinal response alternative for this item
        """
        return [c_i.sum(axis=1).to_numpy()
                for c_i in self.response_counts()]

    def item_total_count(self):
        """Total number of response counts for each item,
        summed across all response alternatives and all groups.
        :return: c = list of total response counts,
            c[i] = scalar integer for i-th item
        """
        return [c_i.sum().sum()
                for c_i in self.response_counts()]


# -------------------------------------------------------- Module TEST:
if __name__ == '__main__':
    from ItemResponseCalc.ir_source import item_response_table, Tables
    from sqlalchemy import create_engine

    HAQ_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry/IRT-IOI-HA/IOI-HA-Data'
    # ioiha_data_path = HAQ_PATH / 'IRT-IOI-HA' / 'IOI-HA-data'
    # to ioi-ha data sets OTHER THAN the Swedish Quality Registry

    print('*** Test Load Questionnaire:\n')
    ioiha_q = Questionnaire.load(HAQ_PATH / 'IOI-HA-English.txt')
    print(ioiha_q)
    # ------------------------------------------------------------------

    print('\n*** Testing Table from Excel with Hickson data set:\n')
    work_path = HAQ_PATH
    test_file = HAQ_PATH / 'Hickson' / 'Short_Eartrak AUS.xlsx'

    irf = item_response_table(source=test_file,
                              items={f'Q0{i}': range(1, 6)
                                     for i in range(1, 8)},
                              index_col=0  # integer for read_excel, but string also seems to work
                              )
    print('printing a few records:')
    for (i, r) in enumerate(irf):
        print(r)
        if i > 10:
            break
    print('printing first few records again:')
    for (i,r) in enumerate(irf):
        print(r)
        if i > 5:
            break

    all_r = [r for r in irf]
    print('number of records=', len(all_r))
    print(f'len(irf) = ', len(irf))

    print('\n*** Test item_response_table using SQLAlchemy engine ***')
    sql_path = work_path / 'test_sql'
    engine = create_engine("sqlite+pysqlite:///" + str(sql_path / 'testHickson.db'))
    ir_sql = item_response_table(source=engine,
                                 sql='ioiha',  # *** "SELECT * FROM ioiha",
                                 index_col='EarRowId',
                                 items={f'Q0{item}': [str(i) for i in range(1, 6)]
                                        for item in range(1, 8)},
                                 )  # NOTE: string-encoded integers here

    # ---------------------------------------------------- Test Hickson in data set:
    print('\n*** Testing using Hickson file(s) in an ItemResponseDataSet object:\n')

    ids = ItemResponseDataSet(questionnaire=ioiha_q, #  Questionnaire(item_response_levels=[5,5,5,5,5,5,5]),
                              groups={'HicksonEarTrak': irf,
                                      'Hickson_sql': ir_sql,
                                      'Hickson_joined': Tables(irf, ir_sql)})

    print(ids)
    print('\nResponse Counts:')
    for (i, (q_i, c_df)) in enumerate(zip(ids.questionnaire.items,
                                          ids.response_counts())
                                      ):
        print(f'\nQuestion: ' + q_i.question)
        print(c_df)

    print('\nitem_response_count=\n', ids.item_response_count())
    # all items have same n_response_levels:
    print('item_total_counts: ', ids.item_total_count())



"""This module implements a Bayesian model of subjects' responses
to items in a questionnaire, using Item Response Theory (IRT).

The present version allows two variants of IRT models for ordinal data:
ir_graded_response_prob.GradedResponse(latent_class),
    with latent_class = either ir_latent.LatentLogistic (default) or .LatentNormal

A future version may implement the Partial Credits model as well.

NOTE: To make the model identifiable, some restriction is necessary.
The zero point on the trait and threshold scale is otherwise arbitrary,
if both latent-variable locations and response thresholds are freely variable.

This behavior is user-controlled by initialization arguments
restrict_trait and restrict_threshold.
The variance of response thresholds among groups is also restricted
by an informtive prior that may be user-controlled by
argument threshold_pseudo_factor to initialize method.

The model is also slightly restricted by a weakly informative global_pop_prior,
for numerical stability in case of extreme response patterns,
e.g., aLL ratings in the highest ordinal category.


*** Main Classes:

OrdinalItemResponseModel, defining a complete Bayesian model with all parameters
    adapted to a data set with individual responses to each item in a questionnaire.
    obtained from subjects in one or more different groups.

*** Usage Example: see template script run_irt.py

*** Reference:
* For Version 1.0.0:
Leijon: Analysis of Ordinal Response Data using Bayesian Item Response Theory package ItemResponseCalc.
Technical report, 2023.

* For Version <= 0.6:
Leijon, Kramer et al.:
Analysis of Data from the International Outcome Inventory for Hearing Aids (IOI-HA)
using Bayesian Item Response Theory. 2020, Int J Audiol.
doi: 10.1080/14992027.2020.1813338

** Version history:
* Versioin 1.0.1:
2026-02-25, bug fix OrdinalItemResponseModel.learn: correct time limit incl minutes

* Version 1.0.0:  new model structure, joint distribution of thresholds and traits within each group
2023-08-30, minor cleanup, final testing
2023-08-12, first function version tested

* Version 0.6.0:
2023-07-12, predictive_individual_cov uses only precision_within if only one group
            (because precision_among is then undefined = NaN).

* Version 0.5.0:
2020-06-13, change internal names scales -> items for clarity
2020-06-15, internal changes to allow choice of response_prob models

2019-07-08, first functional version
2019-07-23, set Initial Trait Scale only here,
    for use by both item_scale and item_respondent initialization
2019-08-15, trying new initialization method, no boolean item_trait_map input
2019-08-16, max_hours limit for learn
2019-08-21, using Pool multiprocessing for learn
"""
# *** initialize with n_traits == n_items, then let learning concentrate TraitProb to fewer traits,
#   and prune when possible, during learning, OR after completed learning ?
#
# **** allow string item_id ? ***
# **** allow string trait_id ? ***
# **** allow tied thresholds across several items, like EmaCalc ??? ***

import numpy as np
import datetime as dt
from copy import deepcopy
from collections import Counter
import pandas as pd

from .ir_base import ItemResponseBase
from .ir_global import ItemResponseGlobal
from .ir_group import ResponseGroup
from .item_trait_selector import TraitProb, TraitSelector
from .ir_graded_response_prob import GradedResponse
from .ir_latent import LatentLogistic

# from .ir_group import logger as group_logger
# *** Does NOT work in multiprocessing.Pool child process

from multiprocessing.pool import Pool
import os

import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # test

# ----------------------------------------------
__version__ = '2023-09-06'

INITIAL_TRAIT_SCALE = 3.  # larger?
# = crude assumed initial scale of model trait values
THRESHOLD_PSEUDO_FACTOR = 1.
# default: global pseudo_groups = THRESHOLD_PSEUDO_FACTOR * number of included groups
# controls prior between-population variability of threshold parameters
# ----------------------------------------------

usePool = True
# usePool = False # for TEST


def _pool_size(n):
    """Estimate number of Pool sub-processes
    :param n: total number of independent jobs to share between processes
    :return: n_processes
    """
    # NOTE: cpu_count() returns n of LOGICAL cpu cores.
    return min(os.cpu_count(), n)


# ---------------------------------------------- Main class:
class OrdinalItemResponseModel:
    """A Bayesian IRT model for discrete ORDINAL rating data,
    available as INDIVIDUAL results in a 2D array R with elements
    R[s, i] = ordinal integer-coded response by s-th subject to the i-th item.

    The number of response alternatives may differ among items.
    The i-th item has L_i ordinal response values.

    NOTE: In most modules of this package,
    the math notation assumes that R[s, i] is an integer in range(L_i),
    i.e., it can have integer values 0, 1,..., L_i-1.

    However, in some input data files,
    the recorded responses may be conventionally coded as values 1, 2,..., L_i,
    with missing values encoded as zero.

    All model parameters are treated as random variables.
    The main parameter distributions are represented by equally probable samples.

    This package uses the Graded Response Model (Samejima, 1969, 1997):
    Each rating R_si is determined by ONE latent random Trait variable
    Y_st = a real-valued random variable in the range (-inf, inf).
    There may be one or more separate trait variables,
    each determining the responses to one or more items.

    IFF the i-th item is associated with the t-th trait,
    the latent variable Y_st is assumed to determine the response R_si as
    tau_{i, l} < Y_st <= tau_{i, l+1} => R_si = l,
    where the response thresholds are defined in
    an array (tau_k0, ..., tau_kL) with strictly increasing elements, and
    tau_k0 = -inf; tau_kL = +inf

    The latent random variable Y_st has a logistic (or normal) distribution,
    with location theta_st = a subject-specific TRAIT variable,
    for the t-th trait of the s-th subject.

    The trait location theta_st is
    the individual performance measures to be estimated.

    The response thresholds are determined by a mapping function
    tau_i(eta_i), where eta_i is a 1D parameter vector for the i-th item.

    The thresholds may be either FREE, or RESTRICTED,
    as specified by boolean initialization argument restrict_threshold.
    If restrict_threshold == False, all (L_i - 1) finite thresholds are free parameters.
    If restrict_threshold == True, ONE middle threshold is fixed at zero,
    so there are only (L_i - 2) free threshold values.
    Thus, the parameter vector eta_i has either (L_i - 1) or (L_i - 2) elements.

    Each item response is modelled as determined by ONE trait variable,
    identified by an item_trait_selector.TraitSelector object.
    The association between traits and items is automatically
    estimated by the learning procedure.

    The model is implemented by main properties
    response_prob = an object defining core response probabilities,
        given item-specific and subject-specific sets of parameters.
        The present version allows only ONE response_prob class:
        ir_graded_response_prob.GradedResponse, (using logistic or normal distribution).
        (The Partial Credits response model may be added in a future version.)
    global_pop_prior = an ir_global.ItemResponseGlobal instance,
        kept constant during learning, used as prior for global_pop.
    global_pop = an ir_global.ItemResponseGlobal instance,
        defining learned overall distribution of parameters common for all groups.
    groups = a list with one item_respondents.RespondentGroup instance for each group of subjects.
        The group object includes all recorded responses for all respondents in that group,
        AND models the distribution of individual traits for those respondents,
        AND models the distribution of traits for a RANDOM INDIVIDUAL
            in the (sub)population from which the group respondents were recruited,
        AND models the distribution of MEAN traits in this (sub)population.
    item_trait_sel = a list of TraitSelector objects, one for each item,
        defining the association between the item and the individual traits
    trait_prob = an item_trait_selector.TraitProb object,
        representing the overall probability for each trait to be used for ANY item.
    """
    # --------------------------------------------------------------------
    def __init__(self, response_prob, questionnaire, base,
                 global_pop_prior, global_pop,
                 groups, item_trait_sel, trait_prob):
        """
        :param response_prob: object that calculates log probability for any response
        :param questionnaire: Questionnaire object with info about the items **** needed here ? ********
        :param base: ir_base.ItemResponseBase object defining indexing of model parameters
        :param global_pop_prior: ir_global.ItemResponseGlobal object used as prior for global_pop
        :param global_pop: ir_global.ItemResponseGlobal object defining global_pop model parameters,
            used as prior for distributions of group-specific parameters across groups
        :param groups: list with item_respondents.ResponseGroup elements
        :param item_trait_sel: list of item_trait_selector.TraitSelector instances, one for each item
        :param trait_prob: single TraitProb object, global_pop for all TraitSelector objects
        """
        self.version = __version__
        self.response_prob = response_prob
        self.questionnaire = questionnaire
        self.base = base
        self.global_pop_prior = global_pop_prior
        self.global_pop = global_pop
        self.groups = groups
        self.item_trait_sel = item_trait_sel
        self.trait_prob = trait_prob
        self.log_prob = []

    def __repr__(self):
        return (self.__class__.__name__
                + '(response_prob= ' + repr(self.response_prob) + ','
                + f'\n\t questionnaire= {self.questionnaire.__class__.__name__} '
                + f'object with {self.questionnaire.n_items} items,'
                + f'\n\t base= {self.base.__class__.__name__} '
                + f'\n\t global_pop_prior= {self.global_pop_prior.__class__.__name__} '
                + f'\n\t global_pop= {self.global_pop.__class__.__name__} '
                + '\n\t groups= [\n\t\t'
                + ',\n\t\t'.join(f'{g.__class__.__name__}(name={repr(g.name)}, with {g.n_subjects} respondents)'
                                 for g in self.groups) + '],'
                + f'\n\t item_trait_sel= [{len(self.item_trait_sel)} TraitSelector objects],'
                + f'mapping items to {self.n_traits} latent traits, '
                + f'\n\t trait_prob = {self.trait_prob.__class__.__name__} object, '
                + f'\n\t version= {self.version}'
                + ')')

    @classmethod
    def initialize(cls, data_set,
                   response_prob=None,
                   latent_class=LatentLogistic,
                   n_traits=None,
                   n_samples=50,
                   restrict_traits=True,
                   restrict_threshold=False,
                   trait_scale=INITIAL_TRAIT_SCALE,
                   threshold_pseudo_factor=THRESHOLD_PSEUDO_FACTOR):
        """Initialize a model crudely from given item response data.
        :param data_set: a single ItemResponseDataSet instance,
            including response data from one or more groups of subjects.
        :param response_prob: (optional) object to calculate response probabilities,
            either GradedResponse, (or PartialCredits **** future)
        :param latent_class: (optional) class of latent variable distribution,
            = ir_latent.LatentLogistic or .LatentNormal (only for GradedResponse)
        :param n_traits: (optional) scalar integer max number of effective latent traits
        :param n_samples: (optional) number of samples of parameters (eta, theta) for each group
        :param restrict_traits: (optional) boolean switch;
            True -> global_pop mean trait fixed at zero.
        :param restrict_threshold: (optional) boolean switch;
            True -> one mid threshold fixed at zero.
        :param trait_scale: (optional) scalar initially assumed scale of model trait values
        :param threshold_pseudo_factor: (optional) scalar to define
            pseudo_groups = threshold_pseudo_factor * n_groups
        :return: a cls instance, crudely initialized for given data_source
            to be refined later, by method learn

        Method:
        0: Initialize base = ir_base.ItemResponseBase with indexing scheme, depending on restrict_threshold
        1.1: Initialize global_pop_prior = weakly informative ir_global.ItemResponseGlobal object,
        1.2: Initialize global_pop = copy of global_pop_prior, EXCEPT threshold parameters
            adjusted to informative values, using total item response counts for all items
        2.1: Initialize all group models with individual response data, for traits == items, 
            with sampled individual theta, using initialized thresholds from global_pop.eta_mean
        2.2: Use PCA on these item-specific crude trait samples
        2.3: Transform all item-specific traits to a (sub)space with n_traits dimensions <= n_items.
        """
        n_groups = len(data_set.groups)
        assert n_groups > 0, 'Must have at least one group'
        if response_prob is None:
            response_prob = GradedResponse(latent_class)
        item_response_count = data_set.item_response_count()
        # item_response_count[i] = 1D array of response counts for i-th item, sum across all groups
        # len(item_response_count[i]) = L_i = number of ordinal response alternatives
        n_items = len(item_response_count)
        if n_traits is None:
            n_traits = n_items
        # ---- Create all model components:
        logger.info(f'Initializing model with restrict_traits={restrict_traits}, '
                    + f'restrict_threshold={restrict_threshold}, '
                    + f'threshold_pseudo_factor={threshold_pseudo_factor}.')
        base = ItemResponseBase.initialize(item_response_count, restrict_threshold, response_prob)
        # = object defining indexing scheme for (eta, theta), and some common logprob methods.
        global_pop_prior = ItemResponseGlobal.initialize(base.n_eta(),
                                                         n_traits,
                                                         restrict_traits,
                                                         pseudo_groups=threshold_pseudo_factor * n_groups)
        # = top hierarchical prior for global population model
        global_pop = deepcopy(global_pop_prior)
        # = global population model for (eta, theta) distributions,
        #   used as prior for all included (sub)populations / groups
        global_pop.set_start(item_response_count, base,
                             trait_scale=trait_scale)
        # = crudely modified, to be suitable to start the VI learning procedure
        trait_prob = TraitProb.initialize(n_traits)
        logger.debug(f'Prior TraitProb.alpha = {np.array2string(trait_prob.alpha, precision=4)}')
        trait_sel = [TraitSelector.initialize(n_traits) for _ in range(n_items)]
        # = objects defining item-trait mapping
        # *** put trait_prob and all trait_sel into ONE single class instance ? **********

        # ----------------------------------------- initialize groups with subjects
        groups = [ResponseGroup.initialize_by_item(responses=g_responses,
                                                   trait_scale=trait_scale,
                                                   base=base,
                                                   item_trait_sel=trait_sel,
                                                   global_pop=global_pop,
                                                   name=g_name,
                                                   n_samples=n_samples)
                  for (g_name, g_responses) in data_set.groups.items()]
        # groups[g] = object containing individual response data for each item,
        #   AND a trait-distribution model for the (sub)population represented by the group of respondents,
        #   initially with mapping ONE item <-> ONE trait, i.e., n_traits == n_items
        # trait_cov = np.sum(g.cov_theta()
        #                    for g in groups)
        trait_cov = sum(g.cov_theta()
                        for g in groups)
        transform_matrix = _trait_rotation(trait_cov, n_traits)
        # to project trait locations to subspace with n_traits dimensions <= n_items
        # such that transformed traits are approximately uncorrelated.
        # *** Without transform_traits, the model never gets sparse, so each item <-> one trait. ***
        for g in groups:
            g.transform_traits(transform_matrix)
        # now all model components are ready to start VI learning
        # *** BUT this initial setting -> perhaps too hard initial item_trait_sel.prob
        # *** does not change during learning. Use a PCA model for traits instead?
        return cls(response_prob,
                   data_set.questionnaire,
                   base,
                   global_pop_prior,
                   global_pop,
                   groups,
                   trait_sel,
                   trait_prob)

    @property
    def n_groups(self):
        return len(self.groups)

    @property
    def n_subjects(self):
        return sum(g.n_subjects for g in self.groups)

    @property
    def n_items(self):
        return len(self.item_trait_sel)

    @property
    def n_traits(self):
        return len(self.trait_prob.mean)

    def item_labels(self):  # *** get from questionnaire if defined there ***
        return [f'Q{i+1}' for i in range(self.n_items)]

    # def trait_labels(self):  # ********* ?

    # ------------------------------------------ General VI learn algorithm:
    def learn(self,
              min_iter=10,
              min_step=0.01,
              max_iter=500,
              max_hours=1.,
              max_minutes=0.,
              callback=None):
        """Learn from all observed response data stored in self.groups,
        using Variational Inference (VI).

        This method adapts sampled distributions for all model parameters
        to maximize a lower bound to the total likelihood of the observed data.
        The resulting sequence of lower-bound values is theoretically guaranteed to be non-decreasing,
        but the sampling variability may cause small variations around the non-decreasing trend.
        :param min_iter: (optional) minimum number of learning iterations
        :param min_step: (optional) minimum data log-likelihood improvement,
                 over the latest min_iter iterations,
                 for learning iterations to continue.
        :param max_iter: (optional) maximum number of iterations, regardless of result.
        :param max_hours = (optional) maximal allowed running time, regardless of result.
        :param max_minutes = (optional) maximal allowed running time, regardless of result.
        :param callback: (optional) function to be called after each iteration step.
            If callable, called as callback(self, log_prob)
            where log_prob == scalar last achieved value of VI lower bound
        :return: log_prob = list of log-likelihood values, one for each iteration.
        Result: updated properties of self
        """
        if max_hours == 0. and max_minutes == 0.:
            max_minutes = 30.
        logger.info('Learning the model. Might take some time! '
                    + f'Max {max_hours:.0f} hours + {max_minutes:.0f} minutes.')
        min_iter = np.max([min_iter, 1])
        end_time = dt.datetime.now() + dt.timedelta(hours=max_hours,
                                                    minutes=max_minutes)
        # last allowed time to start new VI iteration
        log_prob = []
        while (len(log_prob) <= min_iter
               or (log_prob[-1] - log_prob[-1 - min_iter] > min_step
                   and (len(log_prob) < max_iter)
                   and (dt.datetime.now() < end_time))):
            log_prob.append(self.adapt())
            if callable(callback):
                callback(self, log_prob[-1])
            logger.info(f'Done {len(log_prob)} iterations. LL={log_prob[-1]:.1f}')
        self.log_prob = np.array(log_prob)
        if dt.datetime.now() >= end_time:
            logger.warning('Learning stopped at time limit, possibly not yet converged')
        if len(log_prob) >= max_iter:
            logger.warning('Learning stopped at max iterations, possibly not yet converged')
        return self.log_prob

    def adapt(self):
        """One adaptation step for all parameters
        using all response data stored in self.groups.
        :return: ll = scalar lower bound to data log-likelihood,
            incl. negative contributions for KLdiv re priors
        """
        ll_items = self._adapt_items()
        ll_global = self.global_pop.adapt(self.groups, self.global_pop_prior)
        if logger.isEnabledFor(logging.DEBUG):
            global_eta_var = self.global_pop.eta.prec.mean_inv
            item_eta_var = [global_eta_var[i]
                            for i in self.base.eta_slices]
            logger.debug('\n' + '\n'.join(f'Item {i+1} tau_mean = '
                                          + np.array_str(tau_i[1:-1],
                                                         precision=3,
                                                         suppress_small=True)
                                          + f'; Item eta variance = '
                                          + np.array_str(eta_var_i,
                                                         precision=3,
                                                         suppress_small=True)
                                          for (i, (tau_i, eta_var_i)) in enumerate(zip(self.item_thresholds_global(),
                                                                                       item_eta_var))
                                          )
                         )
        ll_groups = self._adapt_groups()
        # *** should be done in this order,
        # because item_trait_sel are used in global_pop update
        # and global_pop and items are used in groups update.
        # All KLdiv calculations must use SAME distributions
        # i.e., the current value AFTER adaptations.
        logger.debug(f'll_items={ll_items}; ll_global={ll_global}; ll_groups={ll_groups}')
        return ll_items + ll_global + ll_groups

    def _adapt_items(self):
        """One adaptation step for all item parameters
        using all response data stored in self.groups.
        :return: ll = scalar lower bound to data log-likelihood
        """
        ll = 0.
        # log_resp = np.sum(g.item_trait_logprob()
        #                   for g in self.groups)
        log_resp = sum(g.item_trait_logprob()
                       for g in self.groups)
        logger.debug('item_trait_sel log_resp=\n' + np.array_str(log_resp,
                                                             precision=3,
                                                             suppress_small=True))
        for (item_sel, lr) in zip(self.item_trait_sel, log_resp):
            ll += item_sel.adapt(lr, self.trait_prob)
        ll += self.trait_prob.adapt(self.item_trait_sel)
        i_t_prob = np.array([i.mean for i in self.item_trait_sel])
        logger.debug('item_trait_sel.mean=\n' + np.array_str(i_t_prob,
                                                             precision=3,
                                                             suppress_small=True))
        return ll

    def _adapt_groups(self):
        """One adaptation step for all parameters
        using all response data stored in self.groups.
        :return: ll = scalar lower bound to data log-likelihood
        """
        # logger.info(f'_adapt_groups: id(logger)={id(logger)}; id(group_logger)={id(group_logger)}')
        # adapt_args = (group_logger,)  # *** pass logger as arg to group.adapt? NO does not work!
        adapt_args = ()
        if usePool and self.n_groups > 1:
            n_pool = _pool_size(self.n_groups)
            ch_size = 1
            logger.debug(f'_adapt_groups Pool: n_pool= {n_pool}, ch_size= {ch_size}')
            with Pool(n_pool) as p:
                self.groups = list(p.imap(_adapt, ((g, adapt_args)
                                                   for g in self.groups),
                                          chunksize=ch_size))
        else:
            logger.debug(f'_adapt_groups map, NoPool')
            self.groups = list(map(_adapt, ((g, adapt_args)
                                            for g in self.groups)))
        for g in self.groups:
            logger.debug(f'group {repr(g.name)}: pop_theta.mu.loc= '
                         + np.array2string(g.pop_theta.mu.loc,
                                           precision=3, suppress_small=True)
                         + f'; ll={g.ll:.2f}')
            # ***** must restore all group refs to original base and global params after work in pool copies
            g.base = self.base
            g.global_pop = self.global_pop
            g.item_trait_sel = self.item_trait_sel
        return sum(g.ll for g in self.groups)

    # ------------------------------------------------------- Result Display Functions:
    def item_thresholds_global(self):
        """Global mean thresholds for all items
        :return: list with elements
            tau_i = 1D array of thresholds INCL. extremes at +-inf
        """
        return self.base.tau(self.global_pop.eta.mu.mean)

    def item_thresholds(self, item='Item', response='Response', sample='Sample'):
        """Response thresholds for all items and all groups
        :param item: index name for items in result
        :param response: index name for response category
        :param sample: index name for sample
        :return: pd.DataFrame with thresholds in one column for each group, and
            one row for each (item, sample, response)
            Thresholds are UPPER interval limits for each ordinal response
            EXCL. extremes at +-inf,
            i.e., number of thresholds = n_categories - 1 for each item.

            Sample values are RELATED across items and responses,
            but UNRELATED among groups.
        """
        def tau_group_item(t):
            """Make thresholds DataFrame for ONE item for ONE group
            :param t: 2D threshold array
                t[n, l] = n-th sample of l-th threshold for i-th item
            :return: pd.Series with index labels ('sample', 'Response')
            """
            ind = pd.MultiIndex.from_product([np.arange(t.shape[0]),
                                              np.arange(t.shape[1])],
                                             names=[sample, response])
            return pd.Series(t.reshape((-1,)), index=ind)
        # -----------------------------------------------------------------
        tau_g = {g.name: pd.concat([tau_group_item(tau_gi[:, 1:-1])
                                    for tau_gi in self.base.tau(g.xi)],
                                   keys=self.item_labels(),
                                   names=[item])
                 for g in self.groups}
        return pd.DataFrame.from_dict(tau_g)

    def item_trait_map(self):
        """Mapping with exactly ONE trait corresponding to each item
        :return: itp = 2D boolean array
            itp[i, t] == True <=> i-th item responses were most probably determined by t-th trait
        """
        return np.array([sc.bool for sc in self.item_trait_sel])

    def prune(self):
        """Reduce model complexity by deleting any latent trait variables
        that did not correspond to any item
        :return: None
        Result: modified internal model properties
        """
        it_map = self.item_trait_map()
        keep = np.any(it_map, axis=0)
        logger.info(f'Pruning model to {sum(keep)} active traits out of {len(keep)} initially allowed.')
        self.trait_prob.prune(keep)
        self.global_pop.prune(keep)
        for sc in self.item_trait_sel:
            sc.prune(keep)
        for g in self.groups:
            g.prune(keep)

    # ------------------------------------------------------ Raw Descriptive Data:
    def item_response_count(self):
        """Total frequency of responses at each ordinal level, for each item,
        by subjects in each included group.
        :return: dict with elements (group_key, g_count), where
            g_count is a list of collections.Counter objects,
            g_count[i][r] = number of responses == r for i-th item,
            with r==0 indicating missing response, and
            r = 1, ..., L_i are actual responses.
            *** change to origin-zero response indices ?
        """
        return {g.name: [Counter(sr_i)
                         for sr_i in g.subject_responses]
                for g in self.groups}

    # ------------------------------------------------------ Predictive Models:
    def theta_range(self):
        """Major range of learned theta values across all items and populations,
        suitable for result plots
        :return: tuple (theta_min, theta_max)
        """
        tau = self.item_thresholds_global()
        tau_min = min(tau_i[1] for tau_i in tau)
        tau_max = max(tau_i[-2] for tau_i in tau)
        latent_scale = self.base.rm_class.latent_class.scale
        return tau_min - latent_scale, tau_max + latent_scale

    def item_response_prob(self, th):
        """Prob of any response category as a function of latent trait
        :param th: 1D array with hypothetical trait values (to be used for plots)
        :return: p = list with estimated probability distributions,
            p[i] = 2D array with response probabilities
            p[i][n, l] = P{l-th response | t[n]} for i-th item
            *** return dict with (item_id, item_prob) instead ?
        """
        def response_prob(tau_i):
            """Response probability mass distributions for ONE item
            :param tau_i: 2D array with
                tau_i[n, l] = n-th sample of lower threshold for l-th response for i-th item
            :return: p_i = 2D array for this item
                p[t, l] = prob. of l-th response, given th[t]
            """
            return np.mean(self.base.rm_class.response_prob(tau_i, th),
                           axis=0)  # mean across samples
        # ----------------------------------------------------------------
        eta = np.concatenate([g.eta for g in self.groups], axis=0)
        # = samples of learned eta distribution for all groups / populations, all items
        return [response_prob(tau_i)
                for tau_i in self.base.tau(eta)]

    def item_fisher_information(self, th):
        """Fisher information for all items, as a function of trait values.
        Conventionally plotted as "Item Information Curves" in IRT / Rasch analysis,
        but it is less interesting in the Bayesian framework,
        that automatically generates a measure of the uncertainty.

        Cramer-Rao inequality:
        The variance of an unbiased POINT estimate of a trait  >= 1 / Fisher Information,
        possibly with equality in some simple cases.

        Fisher information is ADDITIVE across items, because
        item responses are assumed conditionally independent, given the latent trait.
        :param th: array of trait values
        :return: F = 2D array with estimated information functions,
            F[i] = 1D array with Fisher information values for given theta
            p[i, n] = Fisher info when th[n] is the TRUE latent trait value.
        """
        def fisher(tau_i):
            """Calculate Fisher info for ONE item
            :param tau_i: 2D array with
                tau_i[m, l] = m-th sample of lower threshold for l-th response for this item
            :return: F_i = 1D array with
                F_i[n] = Fisher info for this item given th[n]
            """
            p = self.base.rm_class.response_prob(tau_i, th)
            # p[m, n, l] = P{Response = l | tau_i[m], th[n]}
            d_lp_dth = self.base.rm_class.d_log_response_prob_d_theta(tau_i, th)
            # d_lp_dth[m, n, l] = partial derivative d ln p[m, n, l] / d th[n]
            return np.mean(np.sum(p * d_lp_dth**2, axis=-1),
                           axis=0)
        # ------------------------------------------------------------
        eta = np.concatenate([g.eta for g in self.groups], axis=0)
        # eta[m] = m-th sample of learned eta array for all groups / populations, all items
        return np.array([fisher(tau_i)
                         for tau_i in self.base.tau(eta)])

    def instrument_mutual_information(self):
        """Mutual Information I(R, Theta)
        between observed response(s) R and latent Trait(s) Theta,
        showing the average amount of information (in bits) about the trait(s) of a RANDOM INDIVIDUAL,
        that is gained from a single record of responses to the given Questionnaire Instrument.

        Quantifies the ability of the Instrument to classify an individual
        into one of several possible categories, i.e.,
        similar to the Person Separation Index (PSI) in the IRT / Rasch tradition.
        Also related to the conventional Reliability measure
        defined as True_population_variance / Total_measure_variance in range (0, 1.),
        where Total_measure_variance = True_population_variance + Error_variance.
        :return: scalar mutual information (in bits) for ONE response record from the instrument
            by ONE random individual from any population included in the model.
            I(R, Theta) = h(Theta) - h(Theta | R), where
            h(X) = - E{log_2 pdf(X)} is differential entropy of a continuous random variable X

        Method:
        h(Theta) estimated from global p(Theta | mu, Lambda)
        h(Theta | R) approximated by average across all individual q(theta_s | r_s),
            which is represented by a set of samples in the learned model.
        """
        n_pseudo = 0.5
        # representing hypothetical participant NOT included in any group
        h_theta = self.global_pop.theta.entropy()
        # = scalar entropy for theta distribution within any included population
        # NOT including the uncertainty of global mean, if freely variable.
        h_theta_group = sum(g.n_subjects * g.entropy_theta_group()
                            for g in self.groups)
        h_theta_group += n_pseudo * h_theta
        h_theta_group /= (self.n_subjects + n_pseudo)
        # = h(theta) estimated as sample average across all included participants in all groups
        logger.debug(f'Global h(theta)= {h_theta:.3f}. Sample-based h(theta)= {h_theta_group:.3f}. (nats)')
        h_theta_given_response = sum(sum(g.entropy_theta_individual())
                                     for g in self.groups)
        # = sum_s h(theta_s | r_s) over all subjects
        h_theta_given_response += n_pseudo * h_theta  # pseudo-participant not seen
        h_theta_given_response /= (self.n_subjects + n_pseudo)
        mi_sampled = h_theta_group - h_theta_given_response  # nats
        mi = h_theta - h_theta_given_response  # nats
        logger.debug(f'mi_sampled= {mi_sampled:.3f}. mi= {mi:.3f} (nats).')
        return mi / np.log(2.)  # bits

    def item_mean_ordinal(self, theta):
        """Expected value of ordinal responses for all items,
        as a function of trait values.
        :param theta: array of trait values
        :return: r = list with expected "responses" on continuous scale
            r[i][...] = sum_l l * P{response = l | theta[...]} for i-th item
            *** return as dict instead ? ***

        NOTE: Since the ordinal responses are NOT on a linear interval scale,
        it is formally MEANINGLESS to take the mean of ordinal responses.
        Therefore, this result should be used only for illustration purpose,
        with caution about the scale non-linearity.

        Nevertheless, the expectation of ordinal responses
        is sometimes used without hesitation in the Rasch research tradition,
        Ref, e.g.: Wright and Masters (1982), Eq. 5.4.1.
        The expected rating plotted vs individual trait value
        is sometimes called Item Characteristic Curve (ICC).
        """
        return [np.dot(p_r, np.arange(p_r.shape[-1]))
                for p_r in self.item_response_prob(theta)]

    # def predictive_individual_var(self):  # *** not needed
    #     return np.diag(self.predictive_individual_cov())

    def predictive_individual_cov(self):
        """Inter-individual Covariance of traits, re. group mean,
        for a RANDOM INDIVIDUAL from any of the included (sub)populations,
        :return: 2D array with mean covariance matrix

        Method: = doc E{inv(Lambda)},
            where Lambda is estimated inter-individual precision matrix
            for theta values in all included (sub)populations / groups
        """
        return self.global_pop.theta.prec.mean_inv


# ------------------------------ module-internal help functions
def _trait_rotation(c, n_traits):
    """Transformation matrix for trait vectors to (sub-)space
    defined by eigenvectors of cov matrix
    :param c: 2D array with covariance matrix for all initial item-defined traits
        summed across all subjects in all groups
        c.shape == (n_items, n_items)
    :return: proj: 2D projection matrix with orthogonal COLUMN vectors
        proj[:, t] = unit direction vector for t-th trait
        proj.shape == (n_items, n_traits)
        where n_traits <= n_items
        The vectors are scaled such that transformation preserves trait variance
        in the new coordinate system,
        so the initial item scale thresholds should still be reasonably valid.
    """
    (e_val, e_vec) = np.linalg.eigh(c)
    mean_var = np.mean(np.diag(c))
    s = np.sign(e_vec[np.argmax(np.abs(e_vec), axis=0), np.arange(e_vec.shape[1])])
    e_vec *= s
    # e_det = np.linalg.det(e_vec)
    proj = e_vec[:, ::-1][:, :n_traits]  # in decreasing e_val order
    new_var = e_val[::-1][:n_traits]  # in same order
    s = np.sqrt(mean_var / new_var)
    # = scale factor to preserve trait variance
    proj *= s  # *********** needed? *****
    # c_proj = proj.T @ c @ proj  # *** for test, should be diagonal
    return proj
# ---------------------------------------------------------------------


# ------------------------------ help function for Pool multitasking:
# used by map or Pool().imap:
def _adapt(task):
    """dispatch call to given object.adapt(...)
    :param task: arguments defining the task:
        task[0] = object whose adapt method is called
        task[1] = positional arguments for adapt method
    :return: returned object from the called adapt method,
        i.e., an updated copy of object task[0]
    """
    # obj = task[0]
    # arg = task[1]
    # logger.debug(f'{obj}.adapt({arg})')
    return task[0].adapt(*task[1])



"""This module implements a multivariate Gaussian distribution of a random vector
and extends the scipy.stats implementations by including Bayesian learning.

*** Classes:
GaussianRV: a trainable Gaussian distribution of a random vector with INDEPENDENT elements,
    defined by a Gaussian mu array, and gamma-distributed precision parameters.

GaussianGivenPrecision: class for the random mu vector of a Gaussian vector,
    with either CORRELATED OR UNCORRELATED elements.


*** Version History:
* Version 1.0.0: new module within ItemResponseCalc, modified from EmaCalc
"""
import numpy as np
import logging
# from scipy.special import gammaln  # , psi

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST


# -------------------------------------------
class GaussianRV:  # *** need subclass for vectors with CORRELATION ? ***
    """Gaussian distribution of a random 1D (row) array with UNCORRELATED elements.
    The probability density function is
    p(x | mu, Lambda) propto prod_d Lambda_d^0.5 exp(- 0.5 (x_d - mu_d)^2 Lambda_d ), where
    mu = (..., mu_d, ...) is the mu array, and
    Lambda=(..., Lambda_d, ...) is the vector of precision values,
        = inverse variance
        = inverse diagonal of covariance matrix.

    To allow Bayesian learning, mu and Lambda are random variables, with
    p(mu, Lambda) = p(pop_theta | Lambda) p(Lambda), where
    p(mu | Lambda) is implemented by a GaussianGivenPrecision instance, and
    p(Lambda) is implemented by either a GammaRV or a WishartRV instance.
    """
    log_2_pi = np.log(2 * np.pi)  # class constant for mean_logpdf calc

    def __init__(self, mu, prec):
        """
        :param mu: GaussianGivenPrecision instance
        :param prec: precision object, either a 1D GammaRV or a 2D WishartRV instance
        """
        self.mu = mu
        self.prec = prec

#     @classmethod
#     def initialize(cls,
#                    loc,
#                    prec_a, prec_b,
#                    learned_weight=0.001):
#         """Create cls instance with default structure
#         :param loc: 1D array-like location vector = mean of mu attribute
#         :param prec_a: scalar or 1D array-like precision gamma shape parameter
#         :param prec_b: 1D array-like precision gamma inverse-scale parameter
#             len(prec_b) == len(loc)
#         :param learned_weight: (optional) scalar effective number of observations
#             learned_weight = 0. gives a non-informative improper global_pop density
#         """
#         prec = PrecisionRV(a=prec_a, b=prec_b)
#         mu = GaussianGivenPrecision(loc, learned_weight, prec)
#         return cls(mu, prec)
#
    def __repr__(self):
        property_sep = ',\n\t'
        return (self.__class__.__name__ + '(\n\t'
                + property_sep.join(f'{k}={repr(v)}'
                                    for (k, v) in vars(self).items())
                + ')')

    @property
    def loc(self):
        return self.mu.loc

    @property
    def size(self):
        return len(self.loc)

    def _prec_mean_log(self):
        """Mean log det(full precision matrix),
        needed for log normalization constant.
        Simplified version for uncorrelated case.
        :return: scalar
        """
        return np.sum(self.prec.mean_log)

    # *********** separate mean_loglikelihood + log_normalization
    # to save computation during sampling that only needs loglikelihood part.

    def mean_logpdf(self, x):
        """E{ ln pdf( x | self ) }, expectation across all parameters of self
        :param x: 1-dim OR M-dim array or array-like list of sample vectors assumed drawn from self
            x[..., :] = ...-th sample row vector
        :return: scalar or array LL, with
            LL[...] = E{ ln pdf( x[..., :] | self ) }  INCL. log normalization constant
            LL.shape == x.shape[:-1]
        """
        x = np.asarray(x)
        if self.mu.learned_weight <= 0.:
            return np.full(x.shape[:-1], -np.inf)
        # z2 = np.dot((x - self.loc) ** 2, self.prec.mu)
        z2 = _square_mahanalobis(x - self.loc, self.prec.mean)
        # = Mahanalobis distance, z2.shape == x.shape[:-1]
        return (- z2 - self.size / self.mu.learned_weight
                + np.sum(self.prec.mean_log) - self.size * self.log_2_pi  # np.log(2 * np.pi)
                ) / 2

    def d_mean_logpdf(self, x):
        """First derivative of self.mean_logpdf(x) w.r.t x
        :param x: 1-dim OR M-dim array or array-like list of sample vectors assumed drawn from self
            x[..., :] = ...-th sample row vector
        :return: array dLL, with
            dLL[..., i] = d E{ ln pdf( x[..., :] | self ) } / d x[..., i]
            dLL.shape == x.shape
        """
        d = np.asarray(x) - self.loc
        return - d * self.prec.mean

    def relative_entropy(self, othr):
        """Kullback-Leibler divergence between self and othr
        :param othr: single instance of same class as self
        :return: scalar KLdiv(q || p) = E_q{ln q(x) / p(x)},
            where q = self and p = othr
        """
        return (self.mu.relative_entropy(othr.mu, self.prec) +
                self.prec.relative_entropy(othr.prec))

    # def predictive(self, rng=None):  # *****************************
    #     """Predictive distribution of random vector, integrated over parameters.
    #     :param rng: (optional) random.Generator object
    #     :return: rv = single StudentRV instance with independent elements
    #
    #     Scalar Student pdf(x) propto (1 + (1/df) (x-m)^2 / scale^2 )^(- (df + 1) / 2)
    #     where scale = sqrt{ (1+beta) self.prec.inv_scale / (beta self.prec.shape)
    #     and Student df = 2* self.prec.shape
    #     See Leijon EmaCalc report Appendix, or Leijon JASA PairedComp paper appendix
    #     """
    #     beta = self.mu.learned_weight
    #     return StudentRV(loc=self.loc,
    #                      scale=np.sqrt(self.prec.b * (1. + beta) / (beta * self.prec.a)),
    #                      df=2 * self.prec.a,
    #                      rng=rng)


# ----------------------------------------------------------------------
class GaussianGivenPrecision:
    """Conditional Gaussian distribution of the mean of a Gaussian random vector,
    given a 1D precision array representing a diagonal precision matrix,
    OR with correlated elements, given a full 2D precision matrix.

    The probability density function is
    p(mu | Lambda) propto  det(beta Lambda)^0.5 exp(- 0.5 Mahanalobis(mu - m, beta Lambda)^2)
    where
    Mahanalobis(x, L)^2 = the square Mahanalobis distance x*T L x
    mu=(..., mu_d, ...)  is a row vector, sample of random vector self
    m=(..., m_d, ...) is the location of self,
    beta is the scalar learned_weight property
    Lambda = (..., Lambda_ij, ....) = 2D square symmetric positive definite precision matrix,
    OR, in case of UNCORRELATED vector,
    Lambda=(..., Lambda_d, ...) is the 1D diagonal of the full matrix
    defined externally in either case.

    Module function _square_mahanalobis(...) handles both cases.
    """
    def __init__(self, loc, learned_weight):
        """Conditional Gaussian vector, given precision matrix
        Associated precision supplied as argument when needed.
        :param loc: location vector
        :param learned_weight: scalar effective number of learning data
        """
        assert np.isscalar(learned_weight), 'learned_weight must be scalar'
        self.loc = np.asarray(loc)
        self.learned_weight = learned_weight

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    @property
    def size(self):
        return len(self.loc)

    @property
    def mean(self):
        return self.loc

    # def adapt(self, x, w, prior):  # *** overridden not needed in ItemResponseCalc
    #     """Update distribution parameters using observed data and global_pop.
    #     :param x: 2D array or array-like list of sample vectors assumed drawn from self
    #     :param prior: global_pop conjugate distribution, same class as self
    #     :param prior: global_pop conjugate distribution, same class as self
    #     :return: None
    #     Result: updated internal parameters of self
    #     """
    #     m = self.loc  # for debug only
    #     self.learned_weight = prior.learned_weight + np.sum(w)
    #     sx = prior.learned_weight * prior.loc + np.dot(w, x)
    #     self.loc = sx / self.learned_weight
    #     d = self.loc - m  # update change in location
    #     if self.learned_weight > 0.5:
    #         logger.debug('comp loc change: '
    #                      + np.array_str(d, precision=3))

    def relative_entropy(self, othr, prec):
        """Kullback-Leibler divergence between self and othr, given precision
        :param othr: single instance of same class as self
        :param prec: single instance of external precision of self,
            either 1D or 2D array
        :return: scalar KLdiv[q || p] = E_q{ln q(x) / p(x)},
            where q = self and p = othr
        """
        d = len(self.loc)
        md = self.loc - othr.loc
        beta_pq_ratio = othr.learned_weight / self.learned_weight
        return (othr.learned_weight * _square_mahanalobis(md, prec.mean)
                + d * (beta_pq_ratio - np.log(beta_pq_ratio) - 1.)
                ) / 2


    # def predictive(self, rng=None):
    #     """Predictive distribution of self, integrated over self.prec
    #     p(pop_theta) = integral p(pop_theta | prec) p(prec) d_prec, where
    #     p(prec) is represented by the PrecisionRV instance self.prec
    #     :param rng: (optional) random.Generator object
    #     :return: rv = single StudentRV instance
    #
    #     Method: see JASA PairedComp paper Appendix
    #     see also Leijon EmaCalc doc report,
    #     re-checked 2022-01-01
    #     """
    #     beta = self.learned_weight
    #     return StudentRV(loc=self.loc,
    #                      scale=np.sqrt(self.prec.b / (self.prec.a * beta)),
    #                      df=2 * self.prec.a,
    #                      rng=rng)


# --------------------------------------- Module help functions
def _square_mahanalobis(d, mean_prec):  # ***** -> static method ?
    """Precision-weighted square of given distance vector(s)
    for a Gaussian vector with either dependent or independent elements.
    :param d: array-like with difference vector(s)
        d[..., :] = ...-sample assumed drawn from a Gaussian distribution
    :param mean_prec: mean precision
        = 2D symmetric array or 1D array with diagonal precision values
        mean_prec.shape[-1] == d.shape[-1]
    :return: z2 = scalar or array
        z2.shape == d.shape[:-1]
    """
    if mean_prec.ndim == 1:  # diagonal precision, no correlations
        return np.dot(d**2, mean_prec)
    else:  # full symmetric precision matrix, with correlation
        return np.einsum('...i, ij, ...j',
                         d, mean_prec, d)


# ------------------------------------------------- TEST
if __name__ == '__main__':
    from ItemResponseCalc.gamma import GammaRV
    from ItemResponseCalc.wishart import WishartRV
    import copy
    from scipy.optimize import approx_fprime, check_grad

    # --------------------------- Test GaussGivenPrecision
    nx = 3
    m = np.zeros(nx)
    mean_prec1 = np.array(np.arange(1, nx + 1))
    mean_prec2 = np.diag(mean_prec1)

    prec1 = GammaRV(a=1., b=1./mean_prec1)
    prec2 = WishartRV(df=1, scale=mean_prec2)

    mu = GaussianGivenPrecision(loc=m + 1., learned_weight=10)
    mu_prior = GaussianGivenPrecision(loc=m, learned_weight=0.1)
    kl_div1= mu.relative_entropy(mu_prior, prec1)
    print('1D prec: kldiv = ', kl_div1)
    kl_div2= mu.relative_entropy(mu_prior, prec2)
    print('2D prec: kldiv = ', kl_div2)

    # --------------------------- Test d_logpdf
    gx = GaussianRV(mu=mu, prec=prec1)  # a Gauss-gamma RV, 1D prec
    test_x = gx.loc + 2.

    def fun(x):
        return gx.mean_logpdf(x)

    def jac(x):
        return gx.d_mean_logpdf(x)

    print('approx gradient = ', approx_fprime(test_x, fun, epsilon=1e-6))
    print('exact  gradient = ', jac(test_x))
    err = check_grad(fun, jac, test_x, epsilon=1e-6)
    print('check_grad err = ', err)





"""This module defines global model parameters,
common for all groups / populations in an ir_model.OrdinalItemResponseModel,
used as hierarchical priors for all group-specific model parameters.

*** Main Classes:
ItemResponseGlobal: container for all threshold and trait parameter distributions,
    for a global population, possibly including several separate (sub)populations,
    each represented by a group of respondents recruited from that (sub)population.
    The global parameter distribution is prior for all (sub)populations.

GlobalTheta:    container for trait-related parameters
GlobalEta:      container for threshold-related parameters

ThetaMean:  conditional Gaussian distribution of global mean theta,
            given average inter-individual precision matrix within groups.
ThetaMeanFixed:  fixed global mean theta = 0.

EtaMean:    conditional Gaussian distribution of global mean eta, given precision array

PrecFactorGroup:    Precision factor for group means re. global mean, called B in doc.
PrecFactorGroup:    Precision factor for global mean, called C in doc.

ThetaPrecision:     Wishart distribution of inter-individual precision matrix,
                    for individual trait vectors.

EtaPrecisionAmongGroups:    gamma distribution of diagonal precision
                            for eta variations between groups.


*** Version History:
* Version 1.0.0: new module for this version
"""
import numpy as np
from scipy.stats import norm
import logging

from .gauss import GaussianRV, GaussianGivenPrecision
from .student import MultiStudentRV
from .gamma import GammaRV
from .wishart import WishartRV

# ------------------------------- Prior hyperparameters
PROB_PSEUDO_COUNT = 0.5
# = Jeffreys prior for Dirichlet-distributed prob-mass estimated from observed counts

THETA_PSEUDO_COUNT = 0.001
# non-informative for ThetaMean prec_factor_group (B in doc) and prec_factor_global (C in doc)

THETA_PRECISION_DELTA = 0.001
# Prior df of within-group Wishart distribution is set at df = D - 1 + THETA_PRECISION_DELTA,
# where df = approximate weight of prior precision re. one individual respondent.
# Wishart distribution is proper only for df > D - 1,
THETA_SCALE_WITHIN = 1.
# = approximate scale of inter-individual theta variations within groups

ETA_PSEUDO_COUNT = 0.01  # called kappa' in doc. slightly informative for EtaMean distribution
# ETA_A_AMONG  # called f' in doc; = pseudo_groups / 2, specified by user, via ir_model.initialize()
ETA_B_AMONG = 0.01  # called g' in doc
# = prior parameters for global eta distribution
# The prior var[eta_gk] between groups is g' / (f' - 1) = g' with f' = 2.
# The prior variance of global mean eta_k between categories is g' / (f' - 1) kappa'
# ----------------------------------------------------------

logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)
# logger.setLevel(logging.DEBUG)  # test


# ----------------------------------------------------------
class ItemResponseGlobal:
    """Container for global (theta, eta) distributions,
    used as priors for all group-specific parameters.
    All model parameters are implemented as random variables.
    """
    def __init__(self, theta, eta):
        """
        :param theta: GlobalTheta instance for all trait-related model parameters
        :param eta: GlobalEta instance with Gaussian distribution of group-specific threshold parameters
        """
        self.theta = theta
        self.eta = eta

    @classmethod
    def initialize(cls, n_eta, n_traits, restrict_traits, pseudo_groups):
        """Create weakly informative distribution of global parameters,
        to be used as constant top prior for the data-trained global parameter distributions,
        which are in turn used as prior distribution for all group-specific model parameters.
        :param n_eta: scalar integer total number of threshold-defining parameters for all items
        :param n_traits: D = scalar integer number of trait parameters
        :param restrict_traits: boolean switch
        :param pseudo_groups: scalar number of prior "groups"
            with all threshold-parameters forced EQUAL TO global mean.
            Larger value -> smaller between-population variations of thresholds.
        :return: a cls instance
        """
        theta_prec = ThetaPrecision.initialize(n_traits)
        if restrict_traits:
            theta_mean = ThetaMeanFixed.initialize(n_traits)
            prec_factor_global = None  # not used, infinite precision of theta_mean
        else:
            theta_mean = ThetaMean.initialize(n_traits, learned_weight=THETA_PSEUDO_COUNT)
            prec_factor_global = PrecFactorGlobal(a=THETA_PSEUDO_COUNT, b=THETA_PSEUDO_COUNT)
        prec_factor_group = PrecFactorGroup(a=THETA_PSEUDO_COUNT, b=THETA_PSEUDO_COUNT)
        theta_param = GlobalTheta(theta_mean,
                                  theta_prec,
                                  prec_factor_global,
                                  prec_factor_group)
        eta_prec = EtaPrecisionAmongGroups.initialize(n_eta, pseudo_groups)
        eta_param = GlobalEta(EtaMean.initialize(n_eta), eta_prec)
        logger.debug(f'Initializing Global Prior with'
                     +f'\n\tPROB_PSEUDO_COUNT={PROB_PSEUDO_COUNT}'
                     +f'\n\tTHETA_PSEUDO_COUNT={THETA_PSEUDO_COUNT}'
                     +f'\n\tTHETA_SCALE_WITHIN={THETA_SCALE_WITHIN}'
                     +f'\n\tETA_PSEUDO_COUNT={ETA_PSEUDO_COUNT}'
                     +f"\n\tpseudo_groups={pseudo_groups}; f'={pseudo_groups / 2}"
                     +f'\n\tETA_B_AMONG={ETA_B_AMONG}'
                     )
        return cls(theta_param, eta_param)

    def set_start(self, item_response_count, base, trait_scale):
        """Initialize self based on total item response counts,
        defining typical response thresholds,
        to be used for initialization of individual traits in all groups.
        Needed BEFORE the start of iterative VI learning procedure.
        :param item_response_count: list of total response counts, sum across all groups
        :param base: ir_base.ItemResponseBase object, already initialized
        :param trait_scale: scalar initial scale of individual trait locations
        :return: None
        Result: updated self.mu.loc, other params left at their prior values
        """
        self.eta.mu.set_initial_loc(item_response_count, base, trait_scale)

    def adapt(self, groups, prior):
        """Adapt global distributions to current group / (sub)populations
        :param groups: list of ir_group.RespondentGroup instances
        :param prior: object of same class as self, with prior distributions
        :return: LL = scalar = - KLdiv(self || prior)
        """
        return self.eta.adapt(groups, prior.eta) + self.theta.adapt(groups, prior.theta)

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        """
        self.theta.prune(keep)
        # self.eta.prune(keep)  # nothing theta-related there

    # def standardize(self, s):
    #     """Standardize by scaling for unity trait variance, as estimated externally
    #     :param s: 1D array with down-scale factors, one for each trait
    #     :return: None
    #     """
    #     self.theta.standardize(s)
    #     self.eta.standardize(s)


class GlobalTheta:
    """Global distributions of all trait-related parameters
    """
    def __init__(self, mu, prec, prec_factor_global, prec_factor_group):
        """
        :param mu: ThetaMean instance; Gaussian global mean of group-specific mean theta, if freely variable
            OR ThetaMeanFixed instance with loc = 0., in case restricted_trait
        :param prec: ThetaPrecision instance,
            = Wishart-distributed inter-individual trait precision within groups
        :param prec_factor_global: scalar gamma-distributed ratio factor, called C in doc report, such that
            precision of mu = prec_factor_global * prec, IFF theta_mean is free random variable
        :param prec_factor_group: scalar gamma-distributed ratio factor, called B in doc report, such that
            precision of group-specific mean = prec_factor_group * prec
        """
        self.mu = mu
        self.prec = prec
        self.prec_factor_global = prec_factor_global
        self.prec_factor_group = prec_factor_group

    def adapt(self, groups, prior):
        """
        :param groups: list of ir_group.RespondentGroup instances
        :param prior: object of same class as self, with prior distribution
        :return: LL = scalar = - KLdiv(self || prior)
        """
        ll_prec = self.prec.adapt(groups, self, prior.prec)
        logger.debug(f'Global Theta.prec.mean_inv = {np.diag(self.prec.mean_inv)}')
        g_pop_mu = [g.pop_theta.mu for g in groups]
        if self.prec_factor_global is None:  # and self.mu is a ThetaMeanFixed object
            ll_mu = 0.
            ll_factor_c = 0.
        else:
            ll_mu = self.mu.adapt(g_pop_mu, self, prior.mu)
            logger.debug(f'Global Theta.mu.mean = {self.mu.mean}')
            ll_factor_c = self.prec_factor_global.adapt(self, prior.prec_factor_global)
            logger.debug(f'Global <C> = {self.prec_factor_global.mean}')
        ll_factor_b = self.prec_factor_group.adapt(g_pop_mu, self, prior.prec_factor_group)
        logger.debug(f'Global <B> = {self.prec_factor_group.mean}')
        return ll_prec + ll_mu + ll_factor_b + ll_factor_c

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        """
        self.mu.prune(keep)
        self.prec.prune(keep)

    def predictive_mean(self):
        return self.mu.predictive(self.prec)

    def entropy(self):  # **** check! over-estimation ?
        """Mean entropy of the global distribution of trait vector theta
        for a random individual in any included population.
        Expectation is across the variability of learned self.prec
        :return: scalar h = E{ - ln p(theta | self) }

        Method: doc sec:MI eq:entropyGlobalTheta
        """
        d = self.mu.size
        return (d * np.log(2 * np.pi)  + d - self.prec.mean_log_det) / 2


# -------------------------------------- internal parameter classes
class ThetaMean(GaussianGivenPrecision):
    """Implements conditional Gaussian mean of trait locations,
    for the global population, represented by ALL included groups,
    given a precision matrix supplied as argument when needed.
    """
    @classmethod
    def initialize(cls, n_traits, learned_weight):
        """
        :param n_traits: number of traits in model
        :param learned_weight: weight for external precision object
            = effective number of observed data used for learning
        :return: a cls instance
        """
        return cls(loc=np.zeros(n_traits), learned_weight=learned_weight)

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        """
        self.loc = self.loc[keep]

    def weighted_mean_mean(self):
        """weigted outer product matrix of self.loc,
        needed for global prec adapt
        :return: m2 = gamma * self.mean(as column) * self.mean(as row)
            self.mean is written \bar m in doc
        """
        return self.learned_weight * self.loc[:, None] * self.loc

    def adapt(self, g_pop_mu, global_theta, prior):
        """VI update
        :param g_pop_mu: list of ir_group.PopulationMeanTheta instances, one for each group
        :param global_theta: a GlobalTheta instance containing self and associated objects
        :param prior: object of same class as self
        :return: ll = - KLdiv(self || prior) after update

        Method: doc sec:PosteriorTheta eq:PosteriorGlobalMuGamma eq:PosteriorGlobalMuMean
        """
        mean_b = global_theta.prec_factor_group.mean  # = <B> in doc
        self.learned_weight = (global_theta.prec_factor_global.mean  #  <C>
                               + mean_b  # <B>
                               * sum(1. - g_mu.global_weight for g_mu in g_pop_mu)
                               )
        self.loc = (prior.loc   # == 0.
                    + sum(g_mu.loc for g_mu in g_pop_mu)
                    * mean_b / self.learned_weight)
        return - self.relative_entropy(prior, global_theta.prec)

    def predictive(self, prec):
        """Predictive distribution, by integrating out prec
        :param prec: precision matrix implemented by a WishartRV instance
        :return: a MultiStudentRV instance that can generate samples
        """
        nu = prec.df
        df = nu + 1 - self.size
        gamma = self.learned_weight
        return MultiStudentRV(df=df,
                              loc=self.loc,
                              sigma=prec.inv_scale / df / gamma)


class ThetaMeanFixed:
    """Implements a Fixed prior mean of trait locations == 0
    """
    learned_weight = np.inf

    def __init__(self, loc):
        self.loc = loc

    @classmethod
    def initialize(cls, n_traits):
        """
        :param n_traits: number of traits in model
        :return: a cls instance
        """
        return cls(np.zeros(n_traits))

    @property
    def size(self):
        return len(self.loc)

    @property
    def mean(self):
        return self.loc

    def weighted_mean_mean(self):
        """mean_square matrix needed for global prec adapt
        """
        d = len(self.loc)
        return np.zeros((d, d))

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        """
        self.loc = self.loc[keep]

    # def adapt(self, group_pop_mu, global_theta, prior):  # *** should never be called ! *****

    def predictive(self, *args, **kwargs):
        """Fake random variable fixed at self.loc = all zero
        """
        return self

    def rvs(self, size, **kwargs):
        """
        :param size: number of fake sample vectors, all == 0.
        :param kwargs: not used
        :return: array with all-zero samples
        """
        return np.zeros((size, len(self.loc)))



# ------------------------------------- Global threshold-defining parameters
class GlobalEta(GaussianRV):
    """Global Gaussian distribution of threshold-defining parameters eta across groups
    with INDEPENDENT elements.
    """
    # def __init__(self, mu, prec):  # *** superclass

    def adapt(self, groups, prior):
        """
        :param groups: list of ir_group.RespondentGroup instances
        :param prior: object of same class as self, with prior distribution
        :return: LL = scalar = - KLdiv(self || prior)

        Method: doc sec:PosteriorEtaPsiGlobal
        """
        # ***** use superclass adapt jointly for mu and prec ? ******************
        eta_samples = [g.eta for g in groups]
        self.mu.adapt([np.mean(eta_g, axis=0)  # mean across samples
                       for eta_g in eta_samples], prior.mu)
        # logger.debug(f'Global eta.mu.mean = {self.mu.mean}')
        self.prec.adapt([np.mean(eta_g**2, axis=0)
                         for eta_g in eta_samples], prior.prec,
                        self.mu, prior.mu)
        # logger.debug(f'Global eta.prec.mean_inv = {self.prec.mean_inv}')
        return - self.relative_entropy(prior)

    # def relative_entropy(self, othr):   # -> superclass
    #     """Kullback-Leibler divergence KLdiv(self || othr)
    #     :param othr: instance of same class as self
    #     :return: E_self{ln pdf(x | self) / pdf(x | othr) }
    #     """
    #     return (self.mu.relative_entropy(othr.mu, self.prec)
    #             + self.prec.relative_entropy(othr.prec))


class EtaMean(GaussianGivenPrecision):
    """Implements conditional Gaussian distribution of group-specific threshold-defining parameters,
    given external precision.
    """
    # def __init__() by superclass

    @classmethod
    def initialize(cls, n_eta):
        """Create weakly informative prior distribution of
        threshold-defining eta parameters for all items.
        :return: a cls instance
        """
        return cls(loc=np.zeros(n_eta), learned_weight=ETA_PSEUDO_COUNT)

    def set_initial_loc(self, ir_count, base, trait_scale):
        """
        :param ir_count: list of 1D arrays with response counts
            ir_count[i][l] = total number of l-th ordinal response to i-th item, across all groups
        :param base: ir_base.ItemResponseBase instance defining eta -> threshold mapping
        :param trait_scale: presumed st.-dev. of inter-individual trait variations
        :return: a cls instance
        """
        def estim_eta(item_count, scale):
            """Estimate threshold values that might yield observed response count,
            assuming that individual latent variables are Gaussian with known scale.
            :param item_count: 1D count array for ONE item
            :param scale: inter-individual scale of latent variable Y
            :return: eta = 1D array of threshold-defining parameters
            """
            f = np.cumsum(item_count + PROB_PSEUDO_COUNT)
            p_cum = f/ f[-1]
            # = cumulative probabilities for latent variable to fall in each response interval
            # p_cum[l] = P(tau[l] < Y <= tau[l+1])
            tau = norm.ppf(p_cum[:-1], scale=scale)
            # EXCL. extreme thresholds -inf, +inf
            eta = base.thr.tau_inv(tau)
            # such that base.thr.tau(eta) = tau, with required restriction
            return eta
        # ---------------------------------------------------------------
        total_scale = np.sqrt(trait_scale**2 + base.rm_class.latent_scale()**2)
        # = presumed inter-individual st.dev. of latent variable
        eta_all = np.concatenate([estim_eta(item_count, total_scale)
                   for item_count in ir_count])
        self.loc = eta_all

    def adapt(self, eta, prior):  # *** use superclass ?
        """
        :param eta: array-like with observed data:
            eta[g] = 1D array with mean of current eta distribution for g-th group
        :param prior: object of same class as self, with prior distribution
        :return: None

        Method: doc sec:PosteriorEtaPsiGlobal
        """
        self.learned_weight = prior.learned_weight + len(eta)
        prior_eta = prior.learned_weight * prior.loc
        self.loc = (prior_eta + np.sum(eta, axis=0)) / self.learned_weight


# --------------------------------------------------- Global Precision Factors
class PrecFactorGroup(GammaRV):
    """Distribution of ratio factor, called B in doc,
    for the precision of group-specific mean trait vectors,
    relative to the global mean.
    """
    # def __init__(self, a, b):  by superclass

    def adapt(self, group_pop_mu, global_theta, prior):
        """
        :param group_pop_mu: iterable with one ir_group.PopulationMeanTheta instance for each group
            representing the mean trait in the population represented by the group
        :param global_theta: owner of self = ir_global.GlobalTheta instance
        :param prior: object with same class as self
        :return: ll = - KLdiv(self || prior)

        Method: doc sec:PosteriorB
        """
        d2 = [g_mu.mean_sq_dev(global_theta.mu, global_theta.prec)
              for g_mu in group_pop_mu]
        # d2[g] = mean square deviation of g-th group mean from global mean
        d = global_theta.mu.size
        self.a = prior.a + len(d2) * d / 2
        self.b = prior.b + sum(d2) / 2
        return - self.relative_entropy(prior)


class PrecFactorGlobal(GammaRV):
    """Distribution of precision factor, called C in doc,
    for the precision of the global mean trait vector,
    implemented by ir_global.GlobalTheta.mu,
    in case this is an object of class ir_global.ThetaMean
    """
    # def __init__(self, a, b):  by superclass

    def adapt(self, global_theta, prior):
        """
        :param global_theta: owner of self, ir_global.GlobalTheta instance
        :param prior: object with same class as self
        :return: ll = - KLdiv(self || prior)

        Method: doc sec:PosteriorC
        """
        d = len(global_theta.mu.loc)
        self.a = prior.a + d / 2
        gamma = global_theta.mu.learned_weight
        m = global_theta.mu.mean
        mean_prec = global_theta.prec.mean   # <Lambda>
        self.b = prior.b + (d / gamma + m @ mean_prec @ m) / 2
        return - self.relative_entropy(prior)  # superclass method


# ---------------------------------------------------- Precision parameters:
class ThetaPrecision(WishartRV):
    """Wishart-distributed Lambda = global inter-individual precision of trait vectors.
    The probability density function is
    p(Lambda) = (1/C) | Lambda|**((nu - p - 1)/2) exp(- trace(inv(V) Lambda)/2), where
        V = symmetric scale matrix,
        p = length of corresponding Gaussian vector
        nu = degree of freedom parameter >= p,
        Lambda.shape == V.shape == (p, p)
        C = |V|**(nu/2) 2**(nu p / 2) Gamma_p(nu/2) is the normalization factor
    """
    @classmethod
    def initialize(cls, n_traits):
        """Create a weakly informative prior distribution
        for the global precision of inter-individual trait vectors,
        common for all groups / populations
        :param n_traits: number of trait elements for each individual
        :return: a cls instance
        """
        nu = n_traits - 1 + THETA_PRECISION_DELTA
        # = minimal degrees-of-freedom to make the distribution just barely proper
        # We need a PROPER prior to be able to calculate relative entropy
        # between posterior and prior models.
        return cls(df=nu, scale=np.eye(n_traits) * THETA_SCALE_WITHIN / nu)

    # def __init__(self, scale, df):  # superclass

    def adapt(self, groups, global_theta, prior):
        """Update distribution parameters using trait data from all groups and global distributions.
        :param groups: iterable of ir_group.ItemResponseGroup instances
        :param global_theta: ir_global.GlobalTheta instance containing self and other related objects
        :param prior: object of same class as self
        :return: - KLdiv(self || prior)
        Result: updated internal parameters of self

        Method: doc sec:PosteriorLambda, eqs eq:PosteriorNu, eq:PosteriorV
        """
        n_s = 0     # accumulator for number of respondents
        cov = np.zeros(self.scale.shape)  # sum inter-individual covariance
        for g in groups:
            n_s += g.n_subjects
            cov += g.sum_theta_theta()
            mean_theta_g = g.pop_theta.mu.loc
            cov -= g.pop_theta.mu.learned_weight * mean_theta_g[:, None] * mean_theta_g   # eq:PosteriorV
        cov -= global_theta.mu.weighted_mean_mean()  # eq:PosteriorV
        self.scale = np.linalg.inv(prior.inv_scale + cov)   # eq:PosteriorV
        self.df = prior.df + n_s    # eqs eq:PosteriorNu
        return - self.relative_entropy(prior)

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.scale reduced
        """
        self.scale = self.scale[keep, :][:, keep]


# -------------------------------------------------------------------------
class EtaPrecisionAmongGroups(GammaRV):
    """Array of independent gamma distributions for model parameter psi,
    representing diagonal precision matrix of the
    threshold-defining parameter vectors, called eta in doc,
    among different ItemResponseGroup instances
    """
    @classmethod
    def initialize(cls, n_eta, pseudo_groups):
        """Prior global distribution of precision of threshold-defining
        eta parameters across groups
        :param n_eta: total number of eta parameters for all items
        :param pseudo_groups: scalar number of prior "groups"
            with ZERO deviations from global mean eta parameters.
            Large value -> smaller threshold variations between real populations.
        :return: a cls instance
        """
        return cls(a=pseudo_groups / 2, b=ETA_B_AMONG * np.ones(n_eta))

    # def __init__(self, a=1., b=1.):  # ******* use superclass  ********

    def adapt(self, eta2, prior, prior_mu, eta_mu):
        """Adapt to given data
        :param eta2: 2D array-like with square observed mean_eta vectors
            eta2[g][k] = k-th mean squared eta value in g-th group
        :param prior: object of same class as self
        :param prior_mu: EtaMean object associated with prior
        :param eta_mu: EtaMean object associated with self
        :return: ll = KLdiv(self || global_pop) after adjustment

        Method: doc sec:PosteriorEtaPsiGlobal eq:PosteriorPsiShape eq:PosteriorPsiScale
        """
        self.a = prior.a + len(eta2) / 2    # eq:PosteriorPsiShape
        z2 = np.sum(eta2, axis=0) # sum across groups
        z2 += prior_mu.learned_weight * prior_mu.mean**2
        z2 -= eta_mu.learned_weight * eta_mu.mean**2
        self.b = prior.b + z2 / 2   # eq:PosteriorPsiScale
        return - self.relative_entropy(prior)

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.b reduced, self.a is scalar, unchanged
        """
        self.b = self.b[keep]



"""Module defining a Bayesian IRT model for the responses of a group of subjects
to a questionnaire with several items.
Each item response is assumed to be determined by the outcome of a
latent random variable with a subject-specific location.

Each uni-dimensional subject trait may determine responses to one or several items.

*** Main Classes:

ResponseGroup:  model for a group of subjects for which
                response data are available for each item for each subject.

PopulationTheta:    distribution of individual traits for a RANDOM INDIVIDUAL
                    in the (sub)population represented by a group of respondents,
                    used as prior for traits of all respondents in this group
PopulationMeanTheta:    distribution of MEAN traits in this (sub)population

** Version history:
* Version 1.0.0: New model structure
2023-07-19, New ResponseGroup implements both eta and theta parameter distributions, jointly sampled.
            New PopulationMeanTheta is conditional on ir_global.ItemResponseGlobal

* Version 0.6.0:
2023-06-24, adapted for new ir_source, all responses already encoded with origin=0.

* Version 0.5.0 and earlier:
2019-07-08, first version
2019-07-23, cannot use SummaryResponseGroup!
2019-08-11, delete class SummaryResponseGroup
2019-08-11, use items.trait_sel.mean as item_trait_weight
2019-08-15, new initialize method, incl. theta randomization
2019-08-17, single ResponseGroup class, generalized for tied or non-tied prec_among
2019-08-20, ResponseGroup.adapt returns self, to allow multiprocessing
2020-06-15, internal changes to use external response_prob for core functions
"""
# *** with class PopulationMeanTheta for the mean of individual theta
#   and sampling method to get predictive RandomIndividual results?

import numpy as np
from scipy.stats import norm

from samppy import hamiltonian_sampler as ham
from samppy.sample_entropy import entropy_nn_approx as entropy
from ItemResponseCalc.gauss import GaussianGivenPrecision
from ItemResponseCalc.student import MultiStudentGivenMean

import multiprocessing
import logging
logger = logging.getLogger(__name__)

# NOTE: logger does NOT inherit parent handlers, when this module is running as child process
if multiprocessing.current_process().name != "MainProcess":
    # 2023-09-01: Tried to send parent logger as argument: does NOT work.
    # restore a formatter like ir_logging, but only for console output
    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter('{asctime} {name}: {message}',
                                           style='{',
                                           datefmt='%H:%M:%S'))
    logger.addHandler(console)
    logger.setLevel(logging.INFO)
# logger.setLevel(logging.DEBUG)  # TEST

RNG = np.random.default_rng()
# = module-global random generator, Numpy 1.17

ham.VECTOR_AXIS = 1  # just in case default has changed...
# ham.logger.setLevel(logging.DEBUG)  # TEST

ETA_DITHER = 0.2    # scale of random dither for initial eta parameters

# ---------------------------------------------- Respondent Group Model:
class ResponseGroup:
    """Model representing individual distributions latent trait distributions
    and group-common response-thresholds, for ONE group of respondents,
    in an OrdinalItemResponseModel.
    Model is learned by adaptation to individual item responses for each respondent.

    An included population model specifies the distribution of traits
    in the (sub)population from which the respondents were recruited.
    This model is used as prior for individual trait distributions.

    An external global population model represents parameters common for all groups.

    Individual traits theta and group-common threshold-mapping parameters eta are sampled jointly.
    An array xi = self._sampler.x includes all samples of both kinds,
    with a predefined indexing scheme:
    Individual trait samples:
        theta = xi[:, self.base.theta_slice].reshape((n_samples, n_subjects, n_traits))
    Group-specific threshold parameters:
        eta = xi[:, self.base.eta_all_slice], with item-specific sub-arrays
        eta_i = eta[:, self.base.eta_slices[i]] for the i-th item.
    Actual thresholds are mapped by
    tau_i = self.base.thr.tau(eta_i)
    """
    def __init__(self, responses, xi, pop_theta, name, base, item_trait_sel, global_pop):
        """
        :param responses: 2D array-like list with response data
            response[i, s] = integer ordinal response index for i-th item by s-th subject
            encoded with origin=0 and missing response == -1,
        :param xi: 2D array with initial joint (eta, theta) samples, stored as
            xi = concatenated (eta, theta.reshape(...))
            where
            eta[n, :] = n-th sample of all threshold-defining parameters for this group
            theta[n, s, d] = n-th sample of d-th trait for s-th subject
        :param pop_theta: PopulationTheta instance representing current distribution of theta
            in the population represented by this group.
            Used as prior for individual theta distributions of respondents.
        :param name: string with arbitrary label for this object, for display purpose
        :param base: ref to common ItemResponseBase object defining indexing scheme for parameter subsets,
            same for all groups
        :param item_trait_sel: ref to common list of ItemSelector objects
        :param global_pop: ref to common ItemResponseGlobal object
            defining global distributions for all group-specific parameters.

        NOTE: Internal property
        responses = 2D array of individual response indices, with
            response[i, s] = integer index of response to i-th item,
            encoded with origin==0, and missing response encoded as -1.
        The corresponding external property
        subject_responses = responses + 1,
        with origin == 1 and missing response == 0,
        is needed only for some conventional external presentations.
        """
        self.responses = np.array(responses)
        self.n_traits = self.responses.shape[0]
        # NOTE: initially, each item <-> one trait; n_traits may be reduced later
        self.sampler = ham.HamiltonianSampler(fun=self.neg_logprob,
                                              jac=self.d_neg_logprob,
                                              x=xi,  # theta.reshape(n_samples, -1),
                                              epsilon=0.3,
                                              n_leapfrog_steps=10)
        # NOTE: xi = (eta, theta) is stored only as sampler.x
        # args for neg_logprob and d_neg_logprob are defined later, in each call to adapt
        self.pop_theta = pop_theta
        self.name = name
        self.ll = None  # log likelihood to be saved here by adapt step
        self.base = base
        self.item_trait_sel = item_trait_sel
        self.global_pop = global_pop

    def __repr__(self):
        return (self.__class__.__name__ + f'(name= {repr(self.name)},'
                + f'\n\t responses= array with responses to {len(self.responses)} items'
                + f' by {self.n_subjects} respondents,'
                + f'\n\t theta= array with shape= {self.theta.shape},'
                + '\n\t pop_theta= ' + self.pop_theta.__repr__()
                + ')')

    # ------------- initialization methods:
    @classmethod
    def initialize_by_item(cls, responses, trait_scale, base, global_pop, item_trait_sel, name, n_samples=50):
        """Initialize trait values by sampling, one trait for each item
        :param responses: iterable yielding subject responses as a list with
            responses[s][i] = integer index of response to i-th item,
            encoded with origin==0 for the first response alternative,
            and missing response encoded as -1.
        :param trait_scale: scalar initially assumed scale of trait values
        :param base: ItemResponseBase instance created by caller
        :param global_pop: ir_global.ItemResponseGlobal instance
            containing global priors for all group-specific parameters
        :param item_trait_sel: ref to common list of ItemSelector objects
        :param name: string identifying this group
        :param n_samples: number of samples of the intra-subject distribution of trait values
            and intra-group threshold-determining eta parameters
        :return: single cls instance
        """
        # collect all response data in one array
        s_responses = np.array([s_r for s_r in responses]).T  # *** copy needed?
        # s_responses[i, s] = response-level index for i-th item by s-th subject
        # with origin == 0, missing response == -1
        # n_items = len(s_responses)
        eta = np.tile(global_pop.eta.mu.loc, (n_samples, 1))
        eta += ETA_DITHER * RNG.standard_normal(size=eta.shape)
        tau = base.tau(eta)
        # = list of crude initial threshold arrays, INCL extremes in range [-inf, +inf]
        th = np.array([sample_latent(r_i, tau_i, trait_scale=trait_scale)
                       for (r_i, tau_i) in zip(s_responses, tau)])
        # th[i, s, m] = m-th sample for i-th item and s-th subject
        # th = (th + 9. * np.mean(th, axis=0)) / 10.  # *** blur traits across items  ***
        # *** 2023-08-16: did not help to automatically reduce n_traits ***
        th = th.transpose((2, 1, 0))
        # th[m, s, i] = m-th sample of i-th item for s-th subject
        xi = np.concatenate((eta, th.reshape((n_samples, -1))), axis=-1)
        pop_theta = PopulationTheta.initialize(th, global_pop.theta)
        logger.info(f'Initializing {repr(name)}'
                    + f' with {s_responses.shape[-1]} respondent records')
        logger.debug('Initial item pop_theta.loc= ' + np.array2string(pop_theta.mu.loc, precision=3))
        return cls(responses=s_responses,
                   xi=xi,
                   pop_theta=pop_theta,
                   item_trait_sel=item_trait_sel,
                   global_pop=global_pop,
                   name=name,
                   base=base)

    def cov_theta(self):
        """Covariance matrix across all samples, all subjects,
        used only to calculate initial transformation of trait coordinates,
        to be used by method transform_traits later.
        :return: c = 2D symmetric inter-individual covariance matrix
        """
        logger.debug('Initial cov without mean subtraction')
        x = self.theta.reshape((-1, self.theta.shape[-1]))
        c = np.cov(x, rowvar=False)
        return c

    def transform_traits(self, proj):
        """Transform initial item-traits to (sub-)space defined by external PCA
        :param proj: 2D projection matrix with orthonormal COLUMN vectors
            proj[:, d] = unit vector for d-th trait
        :return: None

        Result: self.theta and self.pop_theta transformed to proj coordinate system
        """
        th = np.dot(self.theta, proj)
        self.n_traits = th.shape[-1]
        self.xi = np.concatenate((self.eta, th.reshape((self.n_samples, -1))),
                                 axis=-1)
        self.pop_theta = PopulationTheta.initialize(th, self.global_pop.theta)
        logger.debug(f'Transformed group {repr(self.name)} self.pop_theta.mu.loc= '
                     + np.array2string(self.pop_theta.mu.loc, precision=3, suppress_small=True))

    # ----------------- Methods describing current model:
    @property
    def subject_responses(self):  # ***** needed ?
        """Subject responses as obtained from input files
        using the external encoding {1,..., L_i},
        with missing response == 0
        :return: r = 2D array with responses,
            r[s, i] = response by s-th subject to i-th item
        """
        return self.responses + 1

    @property
    def n_samples(self):
        return self.sampler.x.shape[0]

    @property
    def n_subjects(self):
        return self.responses.shape[-1]

    @property
    def xi(self):
        return self.sampler.x

    @xi.setter
    def xi(self, x):
        self.sampler.x = x

    @property
    def theta(self):
        """Sample representation of current distribution of trait values
        for all subjects in this group
        :return: theta = 3D array of samples
            theta[n, s, d] = n-th sample of d-th trait for s-th subject
        """
        return self.base.theta(self.xi, self.n_subjects)

    @property
    def eta(self):
        """Sample representation of current distribution of eta parameters
        :return: eta = 2D array of samples
            eta[n, k] = n-th sample of k-th eta value, indexed across all items
        """
        return self.base.eta_all(self.xi)

    def sum_theta_theta(self):
        """Sum mean outer product matrix of current theta distribution for each subject,
        averaged across samples, summed across respondents.
        :return: 2D array
            = sum_s E{ theta_s(as column) *  theta_s(as row) },
            with E{} calculated as mean across samples
        """
        th = self.theta
        return np.einsum('nsi, nsj -> ij',
                         th, th) / th.shape[0]

    def entropy_xi(self):
        """Entropy of current (eta, theta) distribution,
        estimated from samples by nearest-neighbor method
        :return: h_sum = scalar entropy
        """
        # h = entropy(self.xi)
        # h = single estimate using all subjects in one operation.
        # Results indicate this gets severely over-estimated ****
        # In previous versions, the distributions were independent among subjects,
        # and Singh et al. (2016) showed convergence is slower with increasing dimension,
        # so an average across all subjects, at low dimension for each subject,
        # might be closer to the true value.

        eta = self.eta
        th = self.theta
        h_eta = entropy(eta)
        h_theta = sum(entropy(th[:, s, :]) for s in range(th.shape[1]))
        # h_theta_test = sum(self.entropy_theta_individual())
        h_sum = h_eta + h_theta
        # as if eta and theta were independent
        # This is theoretically also an over-estimation, but much better than h = entropy(self.xi)
        # logger.debug(f'{self.name}: h_xi = {h}; h_eta= {h_eta}; h_theta= {h_theta}; h_sum= {h_sum}')

        # # *** Try dependent model: p(eta, theta) = p(eta) prod_s p(theta_s | eta):
        # h_eta_theta = sum(entropy(np.concatenate((eta, th[:, s, :]), axis=-1))
        #                   for s in range(th.shape[1]))
        # # = h(eta, theta) with same eta re-used n_subjects times
        # h_sum_xi = h_eta_theta - (self.n_subjects - 1) * h_eta
        # # = another estimate of h(xi) for comparison, *** NO GOOD: large random fluctuations!
        return h_sum

    def entropy_theta_individual(self):
        """Entropy for theta distribution given each participant's response vector
        :return: h_theta_ind = generator yielding
            scalar h(theta_s | r_s) for s-th participant
        """
        th = self.theta.transpose((1, 0, 2))
        # th[s, m, :] = m-th sample of trait vector for s-th subject
        return (entropy(th_s) for th_s in th)

    def entropy_theta_group(self):
        """Entropy for theta distribution among all participants in a group
        :return: scalar h(theta) estimated by ONE sample from each participant
        """
        th = self.theta[0]
        # th[s, :] = ONE random sample of trait vector for s-th subject
        return entropy(th)

    def item_trait_logprob(self):
        """Calculate log-responsibility used to update item-trait selectors
        based on current (eta, theta) distribution defined by self.xi
        :return: lr = 2D array with elements
            lr[i, d] = mean_n{ sum_s ln P(self.responses[i, s] | self.xi[n,:]) } for d-th trait
        """
        return np.mean(self.base.item_response_logprob(self.responses, self.eta, self.theta),
                       axis=1)

    # ----------------- learning methods:
    # def init_adapt(self):  # *** use scipy.minimize for a starting point ? ***

    def adapt(self):
        """One learning step for distribution of group theta values,
        and pop_theta property,
        using response data and current distribution of item scale parameters.
        :return: self, with updated contents, to allow multiprocessing

        Result:
        self.xi = updated set of samples of (eta, theta) for group respondents
        self.pop_theta.mu.loc updated to new self.xi and self.global_pop
        self.ll = E_self{ ln P(self.responses | self) }
                + E{ ln prior_pdf(eta | self.global_pop.eta) }
                + E{ ln prior_pdf(theta | self.pop_theta.mu, global_theta.prec)}
                + entropy(eta, theta) [using samples]
                - KLdiv(self.pop_theta.mu || prior.pop_theta_mu) [given global_theta]
        """
        _id = repr(self.name)  # used for logger output
        logger.debug(self.__class__.__name__ + f'[{_id}].adapt(...)')
        # ---------------------------------- adapt self.xi by Hamilton sampling
        weight = np.asarray([sel.mean for sel in self.item_trait_sel])
        self.sampler.args = (weight,)
        max_steps=20
        # adapt_check_grad(self, weight)  # **** OK. only for test **************
        try:
            self.sampler.safe_sample(min_steps=5, max_steps=max_steps)
            if self.sampler.n_steps >= max_steps:
                logger.warning(f'{_id}: Done {self.sampler.n_steps} = MAX allowed sampling steps')
            else:
                logger.debug(f'{_id}: Done {self.sampler.n_steps} sampling steps')
        except ham.AcceptanceError:  # raised by sampler, even after reducing epsilon
            # logger.debug(f'* AcceptanceError {self}.adapt()')
            # self.sampler.epsilon *= (0.7 + 0.2 * uniform.rvs())  # ************** SKIP
            logger.warning((f'{_id}: AcceptanceError: accept_rate= {self.sampler.accept_rate:.2f} ' +
                            f'of {self.sampler.n_trajectories}; ' +
                            f'epsilon reduced to {self.sampler.epsilon:.5f}'))
            # ****** keep going anyway ********
        else:
            logger.debug(f'{_id}: Sampler accept_rate = {self.sampler.accept_rate:.1%}')
            logger.debug(f'{_id}: Sampler epsilon = {self.sampler.epsilon}')
        # --------------------------------- DONE distribution of theta
        LL = - np.mean(self.sampler.U)
        LL += self.pop_theta.log_normalization(self.global_pop.theta) * self.n_subjects
        # because this log_normalization is not included in self.neg_logprob(...)
        LL += self.pop_theta.mu.adapt(np.mean(self.theta, axis=0), self.global_pop.theta)
        h = self.entropy_xi()
        logger.debug(f'{_id}: LL = {LL:.1f} + entropy={h:.1f}')
        self.ll = LL + h
        logger.debug(self.__class__.__name__ + f'[{_id}].adapt(...) finished')
        return self

    def neg_logprob(self, xi, weight):
        """Negative log-likelihood for any sample subject trait vector
        :param xi: 2D array of sample (eta, theta) values as stored in self.sampler.x
        :param weight: list of 1D arrays of item-trait mapping weights = doc <z_id>
            weight[i][d] = current prob that i-th item response determined by d-th trait
        :return: nlp = 1D vector of negative log-likelihood values
            nlp.shape == (xi.shape[0],) == (self.n_samples,)

        Method: doc eq:LogProbXi, section Respondent- and item-specific parameters
        """
        eta = self.base.eta_all(xi)
        th = self.base.theta(xi, self.n_subjects)
        return (- self.base.response_logprob(self.responses, eta, th, weight)
                - self.global_pop.eta.mean_logpdf(eta)  # INCL. log normalization
                - np.sum(self.pop_theta.mean_loglikelihood(th, self.global_pop.theta),
                         axis=-1)  # EXCL. log normalization
                )

    def d_neg_logprob(self, xi, weight):
        """Gradient of self.neg_logprob for any sample subject trait vector
        :param xi: 2D array of candidate samples (eta, theta) as stored in self.sampler.x
        :param weight: list of 1D arrays of item-trait mapping weights = doc <z_id>
            weight[i][d] = current prob that i-th item response determined by d-th trait
        :return: dnlp = 2D array of gradient values
            dnlp[n, k] = d self.neg_logprob(xi, ...)[n] / d xi[n, k]
            dnlp.shape == xi.shape
        """
        eta = self.base.eta_all(xi)
        th = self.base.theta(xi, self.n_subjects)
        dprior_d_eta = self.global_pop.eta.d_mean_logpdf(eta)
        dprior_d_theta = self.pop_theta.d_mean_loglikelihood(th, self.global_pop.theta)
        # n_samples = dprior_d_theta.shape[0]
        return (- self.base.d_response_logprob(self.responses, eta, th, weight)
                - np.concatenate((dprior_d_eta, dprior_d_theta.reshape(self.n_samples, -1)),
                                 axis=-1)
                )

    # -------------------------------------- Predictive Models for result displays:
    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.theta and self.pop_theta pruned
        """
        th = self.theta[..., keep]
        self.n_traits = np.sum(keep)
        self.xi = np.concatenate((self.eta, th.reshape(self.n_samples, -1)),
                                 axis=-1)
        self.pop_theta.mu.prune(keep)

    def predictive_mean(self):
        """Predictive distribution of MEAN IRT traits
        in the subpopulation for which the subject data in self are representative.
        :return: a student.MultiStudentGivenMean instance
        """
        return self.pop_theta.mu.predictive(self.global_pop.theta)

    def predictive_individual(self):
        """Predictive distribution of IRT traits for a RANDOM INDIVIDUAL
        in the subpopulation for which the respondent data in self are representative.
        :return: a student.MultiStudentGivenMean instance
        """
        return self.pop_theta.predictive(self.global_pop.theta)


# --------------------------------------------------------- help classes:
class PopulationTheta:  # (GaussianRVwithCorrelation):
    """Distribution of individual traits theta in the (sub)population
    from which ONE group of respondents has been recruited.
    Specified by a PopulationMeanTheta instance,
    AND an ir_global.GlobalTheta instance, supplied as argument whenever needed.

    This differs from a standard Bayesian Gaussian random vector,
    because its mean parameter mu is a linear function of the global mean,
    which is another Gaussian variable.
    """
    log_2_pi = np.log(2 * np.pi)  # class constant for mean_logpdf calc

    def __init__(self, mu):
        """
        :param mu: a PopulationMeanTheta instance
        """
        self.mu = mu

    @classmethod
    def initialize(cls, th, global_theta):
        """
        :param th: 3D array of theta samples
            th[n, s, :] = n-th sample of trait vector for s-th respondent
        :param global_theta: a GlobalTheta instance
        :return: a cls instance
        """
        return cls(PopulationMeanTheta(loc=np.mean(th, axis=(0, 1)),
                                       learned_weight=global_theta.prec_factor_group.mean,
                                       global_weight=1.)
                   )

    def mean_loglikelihood(self, th, global_theta):
        """mean log pdf(th | self.mu, global_theta), averaged across self.mu and global_theta,
        EXCL. log normalization constant that remains constant during samplng.
        :param th: array of trait vectors assumed drawn from self
            th[..., :] = ...-th trait vector sample for one or many respondents
        :param global_theta: an ir_global.GlobalTheta instance
        :return: lp = scalar or array
            lp[...] = E{log pdf(th[...] | self, global_theta)} - log(normalization)
            lp.shape == th.shape[:-1]

        Method: doc eq:LogProbXi, section Respondent- and item-specific parameters.
            th is assumed Gaussian, given self.mu and global_theta.prec.
            Result includes ONLY the exponent part, NOT the log det(prec) contribution,
            to save un-necessary computation during sampling.
        """
        th_dev = th - self.mu.conditional_mean(global_theta.mu.loc)
        d2 = np.einsum('...i, ij, ...j',
                       th_dev, global_theta.prec.mean, th_dev)
        # d2[...] = ...-th sample of mean square Mahanalobis distance
        beta_g = self.mu.learned_weight
        gamma = global_theta.mu.learned_weight  # might be inf if global_theta.mu is fixed.
        mean_b = global_theta.prec_factor_group.mean  # <B> in doc
        m2 = (1. + mean_b**2 / beta_g / gamma) * th.shape[-1] / beta_g
        # = contributions from variations in global_theta
        # m2 also constant during sampling, include later *** ?
        return - (d2 + m2) / 2

    def d_mean_loglikelihood(self, th, global_theta):
        """Gradient of mean_loglikelihood(self, th, global_theta) w.t.t. th.
        :param th: 1D or 2D or 3D array of trait vectors assumed drawn from self
            th[..., :] = ...-th sample of trait vector for one or many respondents
        :param global_theta: an ir_global.GlobalTheta instance
        :return: dlp = array
            dlp[..., d] = d logpdf(th, global_theta)[...] / d th[..., d]
            dlp.shape == th.shape
        """
        # th_dev = th - self.mu.conditional_mean(global_theta.mu.loc)
        return - np.dot(th - self.mu.conditional_mean(global_theta.mu.loc),
                        global_theta.prec.mean)

    def log_normalization(self, global_theta):  # **** -> GaussianRVwithCorrelation ? ****
        """log normalization constant, such that
        E{log pdf(th | self, global_theta) }
            = log_normalization(th,...) + log_likelihood(th,...)
            for ONE sample vector th.
        :param global_theta: an ir_global.GlobalTheta instance
        :return: scalar
        """
        return (global_theta.prec.mean_log_det - self.log_2_pi) / 2

    def predictive(self, global_theta):
        """Create a predictive model for the trait distribution
        of a RANDOM INDIVIDUAL in this (sub)population,
        by integrating out the precision.
        The result is conditional on the global_theta.mu,
        as defined by self.mu.conditional_mean(global_mu)
        :param global_theta: an ir_global.GlobalTheta instance
        :return: a MultiStudentGivenMean instance
            that can generate random samples from the predictive distribution
        """
        beta_g = self.mu.learned_weight
        nu = global_theta.prec.df
        df = nu + 1 - self.mu.size
        sigma = global_theta.prec.inv_scale * (beta_g + 1) / beta_g / df
        return MultiStudentGivenMean(mean_factor=self.mu.global_weight,
                                     df=df,
                                     loc=self.mu.loc,
                                     sigma=sigma)

class PopulationMeanTheta(GaussianGivenPrecision):
    """Gaussian conditional distribution of the mean trait vector mu_g
    in the population represented by the respondents in this ResponseGroup instance.
    Defined by super-class properties
        loc = location, part learned from sampled trait vectors theta of ONE group of subjects
        learned_weight = effective number of learning
        AND a scale factor global_weight, for the effect of the global theta mean
    """
    def __init__(self, loc, learned_weight, global_weight=0.):
        super().__init__(loc=loc, learned_weight=learned_weight)
        self.global_weight = global_weight

    @property
    def n_traits(self):  # **** needed ?
        return len(self.loc)

    def conditional_mean(self, global_mu):
        """Conditional mean, given any value of the global mean.
        :param global_mu: any sample value of ir_global.ThetaMean distribution
        :return: 1D array

        Method:
        E{mu_g | global_mu} = (<B> / beta) global_mu + average theta_gs
        doc eq:MuGroupGivenGlobal
        """
        return self.global_weight * global_mu + self.loc

    def mean_sq_dev(self, global_mu, global_prec):  # *** use global_theta as arg ? ***
        """Mean square Mahanalobis distance of self re global mean.
        Needed for VI update of global prec_factor_groups, called B in doc.
        :param global_mu: an ir_global.ThetaMean object
        :param global_prec: global Wishart-distributed precision matrix, called Lambda in doc.
        :return: scalar z2 = E{(mu_g - global_mu) @ Lambda @ (mu_g - global_mu)}
            where
            mu_g = group-specific mean theta represented by self,
            global_mu = global mean theta,
            Lambda = global inter-individual precision matrix

        Method: doc sec:PosteriorB
        """
        d = len(self.loc)
        beta = self.learned_weight
        gamma = global_mu.learned_weight
        # might be inf in case global_mean is fixed at zero
        global_m = global_mu.mean  # called \bar m in doc
        # might be zero in case global_mean is fixed at zero
        d_mean = self.conditional_mean(global_m) - global_m
        return d / beta + (1. + self.global_weight**2) * d / gamma + d_mean @ global_prec.mean @ d_mean

    def adapt(self, mean_theta, global_theta):
        """Adapt self to observed samples and global_pop
        :param mean_theta: 2D array of mean observed samples of subject traits
            mean_theta[s, d] = current mean of d-th trait for s-th subject in this group
        :param global_theta: ir_global.GlobalTheta object with global parameters
        :return: - KLdiv(self || prior)

        Method: doc Sec Population-specific and global trait parameters
            eq:MuGroupGivenGlobal eq:PosteriorBetaGroup eq:PosteriorMuGroup
        """
        mean_b = global_theta.prec_factor_group.mean  # called <B> in doc
        self.learned_weight = len(mean_theta) + mean_b
        self.global_weight = mean_b / self.learned_weight
        self.loc = np.sum(mean_theta, axis=0) / self.learned_weight
        return - self.relative_entropy_re_prior(global_theta)

    def relative_entropy_re_prior(self, global_theta):
        """Mean KLdiv( self.mu_g || prior.mu_g), where
        self.mu_g is Gaussian with
            mean = m(mu) = self.conditional_mean(global_theta.mu),
            prec = beta_g * Lambda = self.learned_weight * global_theta.prec
        prior.mu_g is Gaussian with
            mean = mu = global_theta.mu,
            prec = B * Lambda = global_theta.prec_factor_group * global_theta.prec
        :param global_theta: ir_global.GlobalTheta object with all global parameters
        :return: scalar KL_div = E{ ln pdf(mu_g | self) / pdf(mu_g | prior ) }_{self}

        Method:
        log pdf(mu_g | self) = [ln det(beta_g Lambda) - (mu_g - m(mu)).T beta_g Lambda (mu_g - m(mu)) ] / 2
            = Gaussian with
                mean = m(mu)
                cov = (beta_g Lambda)^{-1}
        log pdf(mu_g | prior)  = [ln det(B Lambda) - (mu_g - mu).T B Lambda (mu_g - mu) ] / 2
            = [ln det(B Lambda) - (mu_g - m(mu) + m(mu) - mu).T B Lambda (mu_g - m(mu) + m(mu) - mu) ] / 2
        m(mu) - mu is Gaussian with
            mean = self.loc + b_ratio * mu.mean - mu.mean
            cov = (1-b_ratio)^2 (gamma Lambda)^{-1}
        """
        d = len(self.loc)  # = self.size
        b_ratio = self.global_weight  # = <B> / beta_g
        b_ratio_m1 = b_ratio - 1.
        mu_mean = global_theta.mu.mean  # might be zero
        mu_dev = self.loc + b_ratio_m1 * mu_mean
        mean_b = global_theta.prec_factor_group.mean  # = <B>
        gamma = global_theta.mu.learned_weight  # might be inf
        return (- np.log(b_ratio)
                + b_ratio_m1
                + b_ratio_m1**2 * mean_b / gamma
                + mu_dev @ global_theta.prec.mean @ mu_dev
                ) * d / 2

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        """
        self.loc = self.loc[keep]

    def predictive(self, global_theta):
        """Create a predictive model for the MEAN trait distribution
        in this (sub)population,
        by integrating out the precision.
        The result is conditional on the global_theta.mu,
        as defined by self.conditional_mean(global_mu)
        :param global_theta: an ir_global.GlobalTheta instance
        :return: a MultiStudentGivenMean instance
            that can generate random samples from the predictive distribution
        """
        beta_g = self.learned_weight
        nu = global_theta.prec.df
        df = nu + 1 - self.size
        sigma = global_theta.prec.inv_scale / beta_g / df
        return MultiStudentGivenMean(mean_factor=self.global_weight,
                                     df=df,
                                     loc=self.loc,
                                     sigma=sigma)


# -------------------------------------------------------- module help functions
def sample_latent(r, tau, trait_scale):
    """Sample individual latent decision variable that might have caused
    observed individual response counts,
    for ONE given item
    :param r: 1D array of response, one from each respondent, for this given item.
        r[s] = response index in range(L_i) for s-th respondent
    :param tau: 2D array of response thresholds, -inf, ..., +inf,
        such that l-th response occurs IFF tau[l] < Y <= tau[l + 1]
        tau.shape[0] == n_samples to generate
    :param trait_scale: scale of inter-individual latent variable
        ***** prior mean of latent var is assumed = 0. *****
    :return: Y = 2D array of samples
        Y[s, n] = n-th sample for s-th respondent,
        random within response interval (tau[r_i[s]], tau[1 + r_i[s]])
    """
    def s_sample(r_s):
        """Sample in transformed (0, 1) range for ONE subject
        :param r_s: scalar integer response by this subject
        :return: 1D array th_s with uniform samples in allowed range
            len(th_s) == tau.shape[0]
        """
        if r_s < 0:
            return RNG.uniform(low=cdf_tau[:, 0], high=cdf_tau[:, -1])
        else:
            return RNG.uniform(low=cdf_tau[:, r_s], high=cdf_tau[:, r_s + 1])

    # ----------------------------------------------------------------------
    cdf_tau = norm.cdf(tau, scale=trait_scale)
    cdf_tau[:, 0] = np.finfo(float).tiny
    # tiny lowest limit, because RANDOM.uniform samples in [0, 1),
    # but we allow only samples in (0., 1)
    cdf_theta = [s_sample(r_s) for r_s in r]
    # cdf_theta[s, n] = n-th sample for s-th respondent
    return norm.ppf(np.array(cdf_theta), scale=trait_scale)


# -------------------------------------------------------- TEST grad:

def adapt_check_grad(gm, weight):  # ******** check eta, theta separately ********
    """Check ItemResponseGroup.d_neg_logprob
    :param gm: ResponseGroup instance
    :param weight: 2D array with item-trait weights
    :return: None
    """
    from scipy.optimize import check_grad, approx_fprime  # only for test

    def fun(x):
        # x = array like gm.xi[0]
        return gm.neg_logprob(gm.xi + x, weight)[0]

    def jac(x):
        # x = array like gm.xi[0]
        return gm.d_neg_logprob(gm.xi + x, weight)[0]

    test_x = np.zeros_like(gm.xi[0])  # like one row of xi
    print(f'*** adapt_check_grad: testing neg_logprob with test_x.shape = {test_x.shape}')
    print('fun =', fun(test_x))
    exact_d = jac(test_x)
    approx_d = approx_fprime(test_x,
                             fun,
                             epsilon=1e-6)
    print('exact jac = ', exact_d)
    err_diff = approx_d - exact_d
    print('approx_d - exact_d = ', err_diff)
    err = check_grad(fun, jac, test_x)
    print('check_grad err = ', err)


# -------------------------------------------------------- TEST:
# if __name__ == '__main__':
    # Testing some module functions OLD version. Replaced by adapt_check_grad
